"""Metrics module.

This module contains classes that implement metrics used to monitor training sessions and evaluate models.
These metrics should all inherit from :class:`thelper.optim.metrics.Metric` to allow them to be dynamically
instantiated by the framework from a configuration file, and evaluated automatically inside a training
session. For more information on this, refer to :class:`thelper.train.trainers.Trainer`.
"""

import copy
import logging
from abc import ABC, abstractmethod
from collections import deque
from typing import Any, AnyStr, Callable, Dict, List, Optional  # noqa: F401

import numpy as np
import sklearn.metrics
import torch
import torch.nn
import torch.nn.functional

import thelper.utils

logger = logging.getLogger(__name__)


class Metric(ABC):
    """Abstract metric interface.

    This interface defines basic functions required so that :class:`thelper.train.trainers.Trainer` can
    figure out how to instantiate, update, reset, and optimize a given metric while training/evaluating
    a model.

    Not all metrics are required to be 'optimizable'; in other words, they do not always need to
    return a scalar value and define a goal. For example, a class can derive from this interface
    and simply accumulate predictions to log them or to produce a graph. In such cases, the class
    would simply need to override the ``goal`` method to return ``None``. Then, the trainer would
    still update the object periodically with predictions, but it will not try to monitor the output
    of its ``eval`` function.
    """

    minimize = float("-inf")
    """Possible return value of the ``goal`` function for scalar metrics."""

    maximize = float("inf")
    """Possible return value of the ``goal`` function for scalar metrics."""

    @abstractmethod
    def accumulate(self, pred, gt, meta=None):
        """Receives the latest prediction and groundtruth tensors from the training session.

        The data given here is used to update the internal state of the metric. For example, a
        classification accuracy metric would accumulate the correct number of predictions in
        comparison to groundtruth labels. The meta values are also provided in case the metric
        can use one of them to produce a better output.

        Args:
            pred: model prediction tensor forwarded by the trainer.
            gt: groundtruth tensor forwarded by the trainer (can be ``None`` if unavailable).
            meta: metadata tensors forwarded by the trainer.
        """
        raise NotImplementedError

    @abstractmethod
    def eval(self):
        """Returns the metric's evaluation result.

        This can be a scalar, a string, or any other type of object. If it is a scalar, the
        metric should also define a goal (minimize, maximize) so that the trainer can monitor
        whether the metric is improving over time. If it is a string, it will be printed in the
        console at the end of every epoch. Otherwise, it is simply logged and eventually returned
        at the end of the training session.
        """
        raise NotImplementedError

    @abstractmethod
    def reset(self):
        """Toggles a reset of the metric's internal state.

        Some metrics might rely on an internal state to smooth out their evaluation results using
        e.g. a moving average, or to keep track of the progression through a dataset. A reset can
        thus be necessary when the dataset changes (e.g. at the end of an epoch). This function is
        called automatically by the trainer in such cases."""
        raise NotImplementedError

    def needs_reset(self):
        """Returns whether the metric needs to be reset between every training epoch or not.

        For example, a metric that computes the prediction accuracy over an entire dataset might
        need to be reset every epoch (it would thus return ``True``). However, if it is implemented
        with a moving average and used to monitor prediction accuracy at each iteration, then
        resetting it every epoch might cause spikes in the results. In this case, returning ``False``
        would be best.

        Note that even if a metric always returns ``False`` here, it might still be reset by the
        trainer if the dataset is switched (e.g. while phasing from training to validation, or to
        testing).
        """
        return True

    @abstractmethod
    def goal(self):
        """Returns the scalar optimization goal of the metric, if available.

        The returned goal can be the ``minimize`` or ``maximize`` members of ``thelper.optim.metrics.Metric``
        if the class's evaluation returns a scalar value, and ``None`` otherwise. The trainer will
        check this value to see if monitoring the metric's evaluation result progression is possible.
        """
        raise NotImplementedError

    def is_scalar(self):
        """Returns whether the metric evaluates to a scalar based on its goal."""
        curr_goal = self.goal()
        return curr_goal == Metric.minimize or curr_goal == Metric.maximize

    def __repr__(self):
        """Returns a generic print-friendly string containing info about this metric."""
        return self.__class__.__module__ + "." + self.__class__.__qualname__ + ": " + str({
            "goal": self.goal(), "is_scalar": self.is_scalar()
        })


class CategoryAccuracy(Metric):
    r"""Classification accuracy metric interface.

    This is a scalar metric used to monitor the multi-label prediction accuracy of a model. By default,
    it works in ``top-k`` mode, meaning that the evaluation result is given by:

    .. math::
      \text{accuracy} = \frac{\text{nb. correct predictions}}{\text{nb. total predictions}} \cdot 100

    When :math:`k>1`, a 'correct' prediction is obtained if any of the model's top :math:`k` predictions
    (i.e. the :math:`k` predictions with the highest score) match the groundtruth label. Otherwise, if
    :math:`k=1`, then only the top prediction is compared to the groundtruth label.

    This metric's goal is to maximize its value :math:`\in [0,100]` (a percentage is returned).

    Usage example inside a session configuration file::

        # ...
        # lists all metrics to instantiate as a dictionary
        "metrics": {
            # ...
            # this is the name of the example metric; it is used for lookup/printing only
            "top_5_accuracy": {
                # this type is used to instantiate the accuracy metric
                "type": "thelper.optim.metrics.CategoryAccuracy",
                # these parameters are passed to the wrapper's constructor
                "params": {
                    # the top prediction count to check for a match with the groundtruth
                    "top_k" 5
                }
            },
            # ...
        }
        # ...

    Attributes:
        top_k: number of top predictions to consider when matching with the groundtruth (default=1).
        max_accum: if using a moving average, this is the window size to use (default=None).
        correct: total number of correct predictions stored using a queue for window-based averaging.
        total: total number of predictions stored using a queue for window-based averaging.
        warned_eval_bad: toggles whether the division-by-zero warning has been flagged or not.
    """

    def __init__(self, top_k=1, max_accum=None):
        """Receives the number of predictions to consider for matches (top-k) and the moving average
        window size (max_accum).

        Note that by default, even if max_accum is not provided here, it can still be set by the
        trainer at runtime through the :func:`thelper.optim.metrics.CategoryAccuracy.set_max_accum` function.
        This is fairly useful as the total size of the training dataset is unlikely to be known when
        metrics are instantiated.
        """
        if not isinstance(top_k, int) or top_k <= 0:
            raise AssertionError("invalid top-k value")
        if max_accum is not None and (not isinstance(max_accum, int) or max_accum <= 0):
            raise AssertionError("invalid max accumulation value for moving average")
        self.top_k = top_k
        self.max_accum = max_accum
        self.correct = deque()
        self.total = deque()
        self.warned_eval_bad = False

    def accumulate(self, pred, gt, meta=None):
        """Receives the latest class prediction and groundtruth labels from the training session.

        The inputs are expected to still be in ``torch.Tensor`` format, but must be located on the
        CPU. This function computes and accumulate the number of correct and total predictions in
        the queues, popping them if the maximum window length is reached.

        Args:
            pred: model class predictions forwarded by the trainer.
            gt: groundtruth labels forwarded by the trainer (can be ``None`` if unavailable).
            meta: metadata forwarded by the trainer (unused).
        """
        if gt is None or gt.numel() == 0:
            return  # only accumulating results when groundtruth available
        if pred.dim() != gt.dim() + 1:
            raise AssertionError("prediction/gt tensors dim mismatch (should be BxCx... and Bx...")
        if pred.shape[0] != gt.shape[0]:
            raise AssertionError("prediction/gt tensors batch size mismatch")
        if pred.dim() > 2 and pred.shape[2:] != gt.shape[1:]:
            raise AssertionError("prediction/gt tensors array size mismatch")
        top_k = pred.topk(self.top_k, dim=1)[1].view(pred.shape[0], self.top_k, -1).numpy()
        true_k = gt.view(gt.shape[0], 1, -1).expand(-1, self.top_k, -1).numpy()
        self.correct.append(np.any(np.equal(top_k, true_k), axis=1).sum(dtype=np.int64))
        self.total.append(gt.numel())
        if self.max_accum and len(self.correct) > self.max_accum:
            self.correct.popleft()
            self.total.popleft()

    def eval(self):
        """Returns the current accuracy (in percentage) based on the accumulated prediction counts.

        Will issue a warning if no predictions have been accumulated yet.
        """
        if len(self.total) == 0 or sum(self.total) == 0:
            if not self.warned_eval_bad:
                self.warned_eval_bad = True
                logger.warning("category accuracy eval result invalid (set as 0.0), no results accumulated")
            return 0.0
        return (float(sum(self.correct)) / float(sum(self.total))) * 100

    def reset(self):
        """Toggles a reset of the metric's internal state, emptying prediction count queues."""
        self.correct = deque()
        self.total = deque()

    def needs_reset(self):
        """If the metric is currently operating in moving average mode, then it does not need to
        be reset (returns ``False``); else returns ``True``."""
        return self.max_accum is None

    def set_max_accum(self, max_accum):
        """Sets the moving average window size.

        This is fairly useful as the total size of the training dataset is unlikely to be known when
        metrics are instantiated. The current implementation of :class:`thelper.train.trainers.Trainer`
        will look for this member function and automatically call it with the dataset size when it is
        available.
        """
        self.max_accum = max_accum

    def goal(self):
        """Returns the scalar optimization goal of this metric (maximization)."""
        return Metric.maximize

    def __repr__(self):
        """Returns a generic print-friendly string containing info about this metric."""
        return self.__class__.__module__ + "." + self.__class__.__qualname__ + ": " + str({
            "top_k": self.top_k
        })


class BinaryAccuracy(Metric):
    r"""Binary classification accuracy metric interface.

    This is a scalar metric used to monitor the binary prediction accuracy of a model. The evaluation
    result is given by:

    .. math::
      \text{accuracy} = \frac{\text{TN} + \text{TP}}{\text{TN} + \text{TP} + \text{FN} + \text{FP}} \cdot 100,

    where TN = True Negative, TP = True Positive, FN = False Negative, and FP = False Positive.

    This metric's goal is to maximize its value :math:`\in [0,100]` (a percentage is returned).

    Usage example inside a session configuration file::

        # ...
        # lists all metrics to instantiate as a dictionary
        "metrics": {
            # ...
            # this is the name of the example metric; it is used for lookup/printing only
            "accuracy": {
                # this type is used to instantiate the accuracy metric
                "type": "thelper.optim.metrics.BinaryAccuracy",
                # there are no useful parameters to give to the constructor
                "params": {}
            },
            # ...
        }
        # ...

    Attributes:
        max_accum: if using a moving average, this is the window size to use (default=None).
        correct: total number of correct predictions stored using a queue for window-based averaging.
        total: total number of predictions stored using a queue for window-based averaging.
        warned_eval_bad: toggles whether the division-by-zero warning has been flagged or not.
    """

    def __init__(self, max_accum=None):
        """Receives the moving average window size (max_accum).

        Note that by default, even if max_accum is not provided here, it can still be set by the
        trainer at runtime through the :func:`thelper.optim.metrics.BinaryAccuracy.set_max_accum` function.
        This is fairly useful as the total size of the training dataset is unlikely to be known when
        metrics are instantiated.
        """
        self.max_accum = max_accum
        self.correct = deque()
        self.total = deque()
        self.warned_eval_bad = False

    def accumulate(self, pred, gt, meta=None):
        """Receives the latest class prediction and groundtruth labels from the training session.

        The inputs are expected to still be in ``torch.Tensor`` format, but must be located on the
        CPU. This function computes and accumulate the number of correct and total predictions in
        the queues, popping them if the maximum window length is reached.

        Args:
            pred: model class predictions forwarded by the trainer.
            gt: groundtruth labels forwarded by the trainer (can be ``None`` if unavailable).
            meta: metadata forwarded by the trainer (unused).
        """
        if gt is None or gt.numel() == 0:
            return  # only accumulating results when groundtruth available
        if pred.dim() != gt.dim() + 1:
            raise AssertionError("prediction/gt tensors dim mismatch (should be BxCx... and Bx...)")
        if pred.shape[0] != gt.shape[0]:
            raise AssertionError("prediction/gt tensors batch size mismatch")
        if pred.dim() > 2 and pred.shape[2:] != gt.shape[1:]:
            raise AssertionError("prediction/gt tensors array size mismatch")
        top = pred.topk(1, dim=1)[1].view(pred.shape[0], 1, -1).numpy()
        true = gt.view(gt.shape[0], 1, -1).numpy()
        self.correct.append(np.any(np.equal(top, true), axis=1).sum(dtype=np.int64))
        self.total.append(gt.numel())
        if self.max_accum and len(self.correct) > self.max_accum:
            self.correct.popleft()
            self.total.popleft()

    def eval(self):
        """Returns the current accuracy (in percentage) based on the accumulated prediction counts.

        Will issue a warning if no predictions have been accumulated yet.
        """
        if len(self.total) == 0 or sum(self.total) == 0:
            if not self.warned_eval_bad:
                self.warned_eval_bad = True
                logger.warning("binary accuracy eval result invalid (set as 0.0), no results accumulated")
            return 0.0
        return (float(sum(self.correct)) / float(sum(self.total))) * 100

    def reset(self):
        """Toggles a reset of the metric's internal state, emptying prediction count queues."""
        self.correct = deque()
        self.total = deque()

    def needs_reset(self):
        """If the metric is currently operating in moving average mode, then it does not need to
        be reset (returns ``False``); else returns ``True``."""
        return self.max_accum is None

    def set_max_accum(self, max_accum):
        """Sets the moving average window size.

        This is fairly useful as the total size of the training dataset is unlikely to be known when
        metrics are instantiated. The current implementation of :class:`thelper.train.trainers.Trainer`
        will look for this member function and automatically call it with the dataset size when it is
        available.
        """
        self.max_accum = max_accum

    def goal(self):
        """Returns the scalar optimization goal of this metric (maximization)."""
        return Metric.maximize


class MeanAbsoluteError(Metric):
    r"""Mean absolute error metric interface.

    This is a scalar metric used to monitor the mean absolute deviation (or error) for a model's
    predictions. This regression metric can be described as:

    .. math::
        e(x, y) = E = \{e_1,\dots,e_N\}^\top, \quad
        e_n = \left| x_n - y_n \right|,

    where :math:`N` is the batch size. If ``reduction`` is not ``'none'``, then:

    .. math::
        MAE(x, y) =
        \begin{cases}
            \operatorname{mean}(E), & \text{if reduction } = \text{mean.}\\
            \operatorname{sum}(E),  & \text{if reduction } = \text{sum.}
        \end{cases}

    `x` and `y` are tensors of arbitrary shapes with a total of `n` elements each.

    Usage example inside a session configuration file::

        # ...
        # lists all metrics to instantiate as a dictionary
        "metrics": {
            # ...
            # this is the name of the example metric; it is used for lookup/printing only
            "mae": {
                # this type is used to instantiate the error metric
                "type": "thelper.optim.metrics.MeanAbsoluteError",
                "params": {
                    "reduction": "mean"
                }
            },
            # ...
        }
        # ...

    Todo: add support for 'dont care' target value?

    Attributes:
        max_accum: if using a moving average, this is the window size to use (default=None).
        reduction: string representing the tensor reduction strategy to use.
        errors: queue of error values stored for window-based averaging.
        warned_eval_bad: toggles whether the division-by-zero warning has been flagged or not.
    """

    def __init__(self, reduction="mean", max_accum=None):
        """Receives the reduction strategy and moving average window size (max_accum).

        Note that by default, even if max_accum is not provided here, it can still be set by the
        trainer at runtime through the :func:`thelper.optim.metrics.BinaryAccuracy.set_max_accum` function.
        This is fairly useful as the total size of the training dataset is unlikely to be known when
        metrics are instantiated.
        """
        self.reduction = reduction
        self.max_accum = max_accum
        self.errors = deque()
        self.warned_eval_bad = False
        if reduction != "mean":
            raise NotImplementedError  # will only be supported with PyTorch version > 1.0.0

    def accumulate(self, pred, target, meta=None):
        """Receives the latest predictions and target values from the training session.

        The inputs are expected to still be in ``torch.Tensor`` format, but must be located on the
        CPU. This function computes and accumulate the error in the queue, popping it if the maximum
        window length is reached.

        Args:
            pred: model prediction values forwarded by the trainer.
            target: target prediction values forwarded by the trainer (can be ``None`` if unavailable).
            meta: metadata forwarded by the trainer (unused).
        """
        if target is None or target.numel() == 0:
            return  # only accumulating results when groundtruth available
        if pred.shape != target.shape:
            raise AssertionError("prediction/gt tensors shape mismatch")
        self.errors.append(torch.nn.functional.l1_loss(pred, target).numpy())  # add reduction here w/ PyTorch v1.0.0
        if self.max_accum and len(self.errors) > self.max_accum:
            self.errors.popleft()

    def eval(self):
        """Returns the current (average) mean absolute error based on the accumulated values.

        Will issue a warning if no predictions have been accumulated yet.
        """
        if len(self.errors) == 0:
            if not self.warned_eval_bad:
                self.warned_eval_bad = True
                logger.warning("mean absolute error eval result invalid (set as 0.0), no results accumulated")
            return 0.0
        return np.mean(np.asarray(list(self.errors)), axis=0, dtype=np.float32)

    def reset(self):
        """Toggles a reset of the metric's internal state, emptying the error queue."""
        self.errors = deque()

    def needs_reset(self):
        """If the metric is currently operating in moving average mode, then it does not need to
        be reset (returns ``False``); else returns ``True``."""
        return self.max_accum is None

    def set_max_accum(self, max_accum):
        """Sets the moving average window size.

        This is fairly useful as the total size of the training dataset is unlikely to be known when
        metrics are instantiated. The current implementation of :class:`thelper.train.trainers.Trainer`
        will look for this member function and automatically call it with the dataset size when it is
        available.
        """
        self.max_accum = max_accum

    def goal(self):
        """Returns the scalar optimization goal of this metric (minimization)."""
        return Metric.minimize


class MeanSquaredError(Metric):
    r"""Mean squared error metric interface.

    This is a scalar metric used to monitor the mean squared deviation (or error) for a model's
    predictions. This regression metric can be described as:

    .. math::
        e(x, y) = E = \{e_1,\dots,e_N\}^\top, \quad
        e_n = \left( x_n - y_n \right)^2,

    where :math:`N` is the batch size. If ``reduction`` is not ``'none'``, then:

    .. math::
        MSE(x, y) =
        \begin{cases}
            \operatorname{mean}(E), & \text{if reduction } = \text{mean.}\\
            \operatorname{sum}(E),  & \text{if reduction } = \text{sum.}
        \end{cases}

    `x` and `y` are tensors of arbitrary shapes with a total of `n` elements each.

    Usage example inside a session configuration file::

        # ...
        # lists all metrics to instantiate as a dictionary
        "metrics": {
            # ...
            # this is the name of the example metric; it is used for lookup/printing only
            "mae": {
                # this type is used to instantiate the error metric
                "type": "thelper.optim.metrics.MeanSquaredError",
                "params": {
                    "reduction": "mean"
                }
            },
            # ...
        }
        # ...

    Todo: add support for 'dont care' target value?

    Attributes:
        max_accum: if using a moving average, this is the window size to use (default=None).
        reduction: string representing the tensor reduction strategy to use.
        errors: queue of error values stored for window-based averaging.
        warned_eval_bad: toggles whether the division-by-zero warning has been flagged or not.
    """

    def __init__(self, reduction="mean", max_accum=None):
        """Receives the reduction strategy and moving average window size (max_accum).

        Note that by default, even if max_accum is not provided here, it can still be set by the
        trainer at runtime through the :func:`thelper.optim.metrics.BinaryAccuracy.set_max_accum` function.
        This is fairly useful as the total size of the training dataset is unlikely to be known when
        metrics are instantiated.
        """
        self.reduction = reduction
        self.max_accum = max_accum
        self.errors = deque()
        self.warned_eval_bad = False
        if reduction != "mean":
            raise NotImplementedError  # will only be supported with PyTorch version > 1.0.0

    def accumulate(self, pred, target, meta=None):
        """Receives the latest predictions and target values from the training session.

        The inputs are expected to still be in ``torch.Tensor`` format, but must be located on the
        CPU. This function computes and accumulate the error in the queue, popping it if the maximum
        window length is reached.

        Args:
            pred: model prediction values forwarded by the trainer.
            target: target prediction values forwarded by the trainer (can be ``None`` if unavailable).
            meta: metadata forwarded by the trainer (unused).
        """
        if target is None or target.numel() == 0:
            return  # only accumulating results when groundtruth available
        if pred.shape != target.shape:
            raise AssertionError("prediction/gt tensors shape mismatch")
        self.errors.append(torch.nn.functional.mse_loss(pred, target).numpy())  # add reduction here w/ PyTorch v1.0.0
        if self.max_accum and len(self.errors) > self.max_accum:
            self.errors.popleft()

    def eval(self):
        """Returns the current (average) mean squared error based on the accumulated values.

        Will issue a warning if no predictions have been accumulated yet.
        """
        if len(self.errors) == 0:
            if not self.warned_eval_bad:
                self.warned_eval_bad = True
                logger.warning("mean squared error eval result invalid (set as 0.0), no results accumulated")
            return 0.0
        return np.mean(np.asarray(list(self.errors)), axis=0, dtype=np.float32)

    def reset(self):
        """Toggles a reset of the metric's internal state, emptying the error queue."""
        self.errors = deque()

    def needs_reset(self):
        """If the metric is currently operating in moving average mode, then it does not need to
        be reset (returns ``False``); else returns ``True``."""
        return self.max_accum is None

    def set_max_accum(self, max_accum):
        """Sets the moving average window size.

        This is fairly useful as the total size of the training dataset is unlikely to be known when
        metrics are instantiated. The current implementation of :class:`thelper.train.trainers.Trainer`
        will look for this member function and automatically call it with the dataset size when it is
        available.
        """
        self.max_accum = max_accum

    def goal(self):
        """Returns the scalar optimization goal of this metric (minimization)."""
        return Metric.minimize


class ExternalMetric(Metric):
    r"""External metric wrapping interface.

    This interface is used to wrap external metrics and use them in the training framework. The metrics
    of ``sklearn.metrics`` are good candidates that have been used extensively with this interface in
    the past, but those of other libraries might also be compatible.

    Along with the name of the class to import and its constructor's parameters, the user must provide
    a handling mode that specifies how prediction and groundtruth data should be handled in this wrapper.
    Also, extra arguments such as target label names, goal information, and window sizes can be provided
    for specific use cases related to the selected handling mode.

    For now, two metric handling modes (both related to classification) are supported:

      * ``classif_best``: the wrapper will accumulate the predicted and groundtruth classification \
        labels forwarded by the trainer and provide them to the external metric for evaluation. If \
        a target label name is specified, then only classifications related to that label will be \
        accumulated. This is the handling mode required for count-based classification metrics such \
        as accuracy, F-Measure, precision, recall, etc.

      * ``classif_score``: the wrapper will accumulate the prediction score of the targeted label \
        along with a boolean that indicates whether this label was the groundtruth label or not. This \
        is the handling mode required for score-based classification metrics such as when computing \
        the area under the ROC curve (AUC).

    Usage examples inside a session configuration file::

        # ...
        # lists all metrics to instantiate as a dictionary
        "metrics": {
            # ...
            # this is the name of the first example metric; it is used for lookup/printing only
            "f1_score_reject": {
                # this type is used to instantiate the wrapper
                "type": "thelper.optim.metrics.ExternalMetric",
                # these parameters are passed to the wrapper's constructor
                "params": {
                    # the external class to import
                    "metric_name": "sklearn.metrics.f1_score",
                    # the parameters passed to the external class's constructor
                    "metric_params": [],
                    # the wrapper metric handling mode
                    "metric_type": "classif_best",
                    # the target class name (note: dataset-specific)
                    "target_name": "reject",
                    # the goal type of the external metric
                    "goal": "max"
                }
            },
            # this is the name of the second example metric; it is used for lookup/printing only
            "roc_auc_accept": {
                # this type is used to instantiate the wrapper
                "type": "thelper.optim.metrics.ExternalMetric",
                # these parameters are passed to the wrapper's constructor
                "params": {
                    # the external class to import
                    "metric_name": "sklearn.metrics.roc_auc_score",
                    # the parameters passed to the external class's constructor
                    "metric_params": [],
                    # the wrapper metric handling mode
                    "metric_type": "classif_score",
                    # the target class name (note: dataset-specific)
                    "target_name": "accept",
                    # the goal type of the external metric
                    "goal": "max"
                }
            },
            # ...
        }
        # ...

    Attributes:
        metric_goal: goal of the external metric, used for monitoring. Can be ``min``, ``max``, or ``None``.
        metric_type: handling mode of the external metric. Can only be one of the predetermined values.
        metric: type of the external metric that will be instantiated when ``eval`` is called.
        metric_params: dictionary of parameters passed to the external metric on instantiation.
        target_name: name of the targeted label. Used only in handling modes related to classification.
        target_idx: index of the targeted label. Used only in handling modes related to classification.
        class_names: holds the list of class label names provided by the dataset parser. If it is not
            provided when the constructor is called, it will be set by the trainer at runtime.
        force_softmax: specifies whether a softmax operation should be applied to the prediction scores
            obtained from the trainer. Only used with the "classif_score" handling mode.
        max_accum: if using a moving average, this is the window size to use (default=None).
        pred: queue used to store predictions-related values for window-based averaging.
        gt: queue used to store groundtruth-related values for window-based averaging.
    """

    def __init__(self, metric_name, metric_type, metric_params=None, target_name=None,
                 goal=None, class_names=None, max_accum=None, force_softmax=True):
        """Receives all necessary arguments for wrapper initialization and external metric instantiation.

        See :class:`thelper.optim.metrics.ExternalMetric` for information on arguments.
        """
        if not isinstance(metric_name, str):
            raise AssertionError("metric_name must be fully qualifiied class name to import")
        if metric_params is not None and not isinstance(metric_params, (list, dict)):
            raise AssertionError("metric_params must be dictionary")
        supported_handling_types = [
            "classif_top1", "classif_best",  # the former is for backwards-compat with the latter
            "classif_scores", "classif_score",  # the former is for backwards-compat with the latter
            "regression",  # missing impl, work in progress
        ]
        if not isinstance(metric_type, str) or metric_type not in supported_handling_types:
            raise AssertionError("unknown metric type '%s'" % str(metric_type))
        if metric_type == "classif_top1":
            metric_type = "classif_best"  # they are identical, just overwrite for backwards compat
        if metric_type == "classif_scores":
            metric_type = "classif_score"  # they are identical, just overwrite for backwards compat
        self.metric_goal = None
        if goal is not None:
            if isinstance(goal, str) and "max" in goal.lower():
                self.metric_goal = Metric.maximize
            elif isinstance(goal, str) and "min" in goal.lower():
                self.metric_goal = Metric.minimize
            else:
                raise AssertionError("unexpected goal type for '%s'" % str(metric_name))
        self.metric_type = metric_type
        self.metric = thelper.utils.import_class(metric_name)
        self.metric_params = metric_params if metric_params is not None else {}
        if "classif" in metric_type:
            self.target_name = target_name
            self.target_idx = None
            self.class_names = None
            if class_names is not None:
                self.set_class_names(class_names)
            if metric_type == "classif_score":
                self.force_softmax = force_softmax  # only useful in this case
        # elif "regression" in metric_type: missing impl for custom handling
        self.max_accum = max_accum
        self.pred = deque()
        self.gt = deque()

    def set_class_names(self, class_names):
        """Sets the class label names that must be predicted by the model.

        This is only useful in metric handling modes related to classification. The goal of having
        class names here is to translate a target class label (provided in the constructor) into a
        target class index. This is required as predictions are not mapped to their original names
        (in string format) before being forwarded to this object by the trainer.

        The current implementation of :class:`thelper.train.trainers.Trainer` will automatically call
        this function at runtime if it is available, and provide the dataset's classes as a list of
        strings.
        """
        if "classif" in self.metric_type:
            if not isinstance(class_names, list):
                raise AssertionError("expected list for class names")
            if len(class_names) < 2:
                raise AssertionError("not enough classes in provided class list")
            if self.target_name is not None:
                if self.target_name not in class_names:
                    raise AssertionError("could not find target name '%s' in class names list" % str(self.target_name))
                self.target_idx = class_names.index(self.target_name)
            self.class_names = class_names
        else:
            raise AssertionError("unexpected class list with metric type other than classif")

    def accumulate(self, pred, gt, meta=None):
        """Receives the latest prediction and groundtruth tensors from the training session.

        The handling of the data received here will depend on the current handling mode.

        Args:
            pred: model prediction tensor forwarded by the trainer.
            gt: groundtruth tensor forwarded by the trainer (can be ``None`` if unavailable).
            meta: metadata tensors forwarded by the trainer.
        """
        if "classif" in self.metric_type:
            if gt is None:
                return  # only accumulating results when groundtruth available
            if self.target_name is not None and self.target_idx is None:
                raise AssertionError("could not map target name '%s' to target idx, missing class list" % self.target_name)
            elif self.target_idx is not None:
                pred_label = pred.topk(1, 1)[1].view(len(gt))
                y_true, y_pred = [], []
                if self.metric_type == "classif_best":
                    must_keep = [y_pred == self.target_idx or y_true == self.target_idx for y_pred, y_true in zip(pred_label, gt)]
                    for idx in range(len(must_keep)):
                        if must_keep[idx]:
                            y_true.append(gt[idx].item() == self.target_idx)
                            y_pred.append(pred_label[idx].item() == self.target_idx)
                else:  # self.metric_type == "classif_score"
                    if self.force_softmax:
                        with torch.no_grad():
                            pred = torch.nn.functional.softmax(pred, dim=1)
                    for idx in range(len(gt)):
                        y_true.append(gt[idx].item() == self.target_idx)
                        y_pred.append(pred[idx, self.target_idx].item())
                self.gt.append(y_true)
                self.pred.append(y_pred)
            else:
                if self.metric_type == "classif_best":
                    self.gt.append([gt[idx].item() for idx in range(len(pred.numel()))])
                    self.pred.append([pred[idx].item() for idx in range(len(pred.numel()))])
                else:  # self.metric_type == "classif_score"
                    raise AssertionError("score-based classification analyses (e.g. roc auc) must specify target label")
        elif self.metric_type == "regression":
            raise NotImplementedError
        else:
            raise AssertionError("unknown metric type '%s'" % str(self.metric_type))
        while self.max_accum and len(self.gt) > self.max_accum:
            self.gt.popleft()
            self.pred.popleft()

    def eval(self):
        """Returns the external metric's evaluation result."""
        if "classif" in self.metric_type:
            y_gt = [gt for gts in self.gt for gt in gts]
            y_pred = [pred for preds in self.pred for pred in preds]
            if len(y_gt) != len(y_pred):
                raise AssertionError("list flattening failed")
            if isinstance(self.metric_params, list):
                return self.metric(y_gt, y_pred, *self.metric_params)
            elif isinstance(self.metric_params, dict):
                return self.metric(y_gt, y_pred, **self.metric_params)
            else:
                return self.metric(y_gt, y_pred, self.metric_params)
        else:
            raise NotImplementedError

    def reset(self):
        """Toggles a reset of the metric's internal state, emptying pred/gt queues."""
        self.gt = deque()
        self.pred = deque()

    def needs_reset(self):
        """If the metric is currently operating in moving average mode, then it does not need to
        be reset (returns ``False``); else returns ``True``."""
        return self.max_accum is None

    def set_max_accum(self, max_accum):
        """Sets the moving average window size.

        This is fairly useful as the total size of the training dataset is unlikely to be known when
        metrics are instantiated. The current implementation of :class:`thelper.train.trainers.Trainer`
        will look for this member function and automatically call it with the dataset size when it is
        available.
        """
        self.max_accum = max_accum

    def goal(self):
        """Returns the scalar optimization goal of this metric (user-defined)."""
        return self.metric_goal


class ClassifReport(Metric):
    """Classification report interface.

    This class provides a simple interface to ``sklearn.metrics.classification_report`` so that all
    count-based metrics can be reported at once under a string-based representation. Note that since
    the evaluation result is a string, this metric cannot be used to directly monitor training
    progression, and thus returns ``None`` in :func:`thelper.optim.metrics.ClassifReport.goal`.

    Usage example inside a session configuration file::

        # ...
        # lists all metrics to instantiate as a dictionary
        "metrics": {
            # ...
            # this is the name of the example metric; it is used for lookup/printing only
            "classifreport": {
                # this type is used to instantiate the classification report metric
                "type": "thelper.optim.metrics.ClassifReport",
                # we do not need to provide any parameters to the constructor, defaults are fine
                "params": {}
            },
            # ...
        }
        # ...

    Attributes:
        report: report generator function, called at evaluation time to generate the output string.
        class_names: holds the list of class label names provided by the dataset parser. If it is not
            provided when the constructor is called, it will be set by the trainer at runtime.
        pred: queue used to store the top-1 (best) predicted class indices at each iteration.
        gt: queue used to store the groundtruth class indices at each iteration.
    """

    def __init__(self, class_names=None, sample_weight=None, digits=4):
        """Receives the optional class names and arguments passed to the report generator function.

        Args:
            class_names: holds the list of class label names provided by the dataset parser. If it is not
                provided when the constructor is called, it will be set by the trainer at runtime.
            sample_weight: sample weights, forwarded to ``sklearn.metrics.classification_report``.
            digits: metrics output digit count, forwarded to ``sklearn.metrics.classification_report``.
        """

        def gen_report(y_true, y_pred, _class_names):
            if not _class_names:
                res = sklearn.metrics.classification_report(y_true, y_pred,
                                                            sample_weight=sample_weight,
                                                            digits=digits)
            else:
                _y_true = [_class_names[classid] for classid in y_true]
                _y_pred = [_class_names[classid] if (0 <= classid < len(_class_names)) else "<unset>" for classid in y_pred]
                res = sklearn.metrics.classification_report(_y_true, _y_pred,
                                                            sample_weight=sample_weight,
                                                            digits=digits)
            return "\n" + res

        self.report = gen_report
        self.class_names = class_names
        if class_names and not isinstance(class_names, list):
            raise AssertionError("expected class names to be list")
        self.pred = None
        self.gt = None

    def set_class_names(self, class_names):
        """Sets the class label names that must be predicted by the model.

        The current implementation of :class:`thelper.train.trainers.Trainer` will automatically
        call this function at runtime if it is available, and provide the dataset's classes as a
        list of strings.
        """
        if class_names and not isinstance(class_names, list):
            raise AssertionError("expected class names to be list")
        if len(class_names) < 2:
            raise AssertionError("class list should have at least two elements")
        self.class_names = class_names

    def accumulate(self, pred, gt, meta=None):
        """Receives the latest class prediction and groundtruth labels from the training session.

        Args:
            pred: model class predictions forwarded by the trainer (in ``torch.Tensor`` format).
            gt: groundtruth labels forwarded by the trainer (can be ``None`` if unavailable).
            meta: metadata forwarded by the trainer (unused).
        """
        if gt is None:
            return  # only accumulating results when groundtruth available
        if self.pred is None:
            self.pred = pred.topk(1, 1)[1].view(len(gt))
            self.gt = gt.view(len(gt)).clone()
        else:
            self.pred = torch.cat((self.pred, pred.topk(1, 1)[1].view(len(gt))), 0)
            self.gt = torch.cat((self.gt, gt.view(len(gt))), 0)

    def eval(self):
        """Returns the classification report as a multi-line print-friendly string."""
        if self.pred is None:
            return "<UNAVAILABLE>"
        return self.report(self.gt.numpy(), self.pred.numpy(), self.class_names)

    def reset(self):
        """Toggles a reset of the metric's internal state, emptying queues."""
        self.pred = None
        self.gt = None

    def goal(self):
        """Returns ``None``, as this class should not be used to directly monitor the training progress."""
        return None


class ConfusionMatrix(Metric):
    """Confusion matrix report interface.

    This class provides a simple interface to ``sklearn.metrics.confusion_matrix`` so that a full
    confusion matrix can be easily reported under a string-based representation. Note that since
    the evaluation result is a string, this metric cannot be used to directly monitor training
    progression, and thus returns ``None`` in :func:`thelper.optim.metrics.ConfusionMatrix.goal`.

    It also offers a tensorboardX-compatible output image that can be saved locally or posted to
    tensorboard for browser-based visualization.

    Usage example inside a session configuration file::

        # ...
        # lists all metrics to instantiate as a dictionary
        "metrics": {
            # ...
            # this is the name of the example metric; it is used for lookup/printing only
            "confmat": {
                # this type is used to instantiate the confusion matrix report metric
                "type": "thelper.optim.metrics.ConfusionMatrix",
                # we do not need to provide any parameters to the constructor, defaults are fine
                "params": {}
            },
            # ...
        }
        # ...

    Attributes:
        matrix: report generator function, called at evaluation time to generate the output string.
        class_names: holds the list of class label names provided by the dataset parser. If it is not
            provided when the constructor is called, it will be set by the trainer at runtime.
        draw_normalized: defines whether rendered confusion matrices should be normalized or not.
        pred: queue used to store the top-1 (best) predicted class indices at each iteration.
        gt: queue used to store the groundtruth class indices at each iteration.
    """

    def __init__(self, class_names=None, draw_normalized=True):
        """Receives the optional class label names used to decorate the output string.

        Args:
            class_names: holds the list of class label names provided by the dataset parser. If it is not
                provided when the constructor is called, it will be set by the trainer at runtime.
            draw_normalized: defines whether rendered confusion matrices should be normalized or not.
        """

        def gen_matrix(y_true, y_pred, _class_names):
            if not _class_names:
                res = sklearn.metrics.confusion_matrix(y_true, y_pred)
            else:
                _y_true = [_class_names[classid] for classid in y_true]
                _y_pred = [_class_names[classid] if (0 <= classid < len(_class_names)) else "<unset>" for classid in y_pred]
                res = sklearn.metrics.confusion_matrix(_y_true, _y_pred, labels=_class_names)
            return res

        self.matrix = gen_matrix
        self.class_names = None
        if class_names is not None:
            self.set_class_names(class_names)
        self.pred = None
        self.gt = None
        self.draw_normalized = draw_normalized

    def set_class_names(self, class_names):
        """Sets the class label names that must be predicted by the model.

        The current implementation of :class:`thelper.train.trainers.Trainer` will automatically
        call this function at runtime if it is available, and provide the dataset's classes as a
        list of strings.
        """
        if not isinstance(class_names, list):
            raise AssertionError("expected class names to be list")
        if len(class_names) < 2:
            raise AssertionError("class list should have at least two elements")
        self.class_names = copy.deepcopy(class_names)
        self.class_names.append("<unset>")

    def accumulate(self, pred, gt, meta=None):
        """Receives the latest class prediction and groundtruth labels from the training session.

        Args:
            pred: model class predictions forwarded by the trainer (in ``torch.Tensor`` format).
            gt: groundtruth labels forwarded by the trainer (can be ``None`` if unavailable).
            meta: metadata forwarded by the trainer (unused).
        """
        if gt is None:
            return  # only accumulating results when groundtruth available
        if self.pred is None:
            self.pred = pred.topk(1, 1)[1].view(len(gt))
            self.gt = gt.view(len(gt)).clone()
        else:
            self.pred = torch.cat((self.pred, pred.topk(1, 1)[1].view(len(gt))), 0)
            self.gt = torch.cat((self.gt, gt.view(len(gt))), 0)

    def eval(self):
        """Returns the confusion matrix as a multi-line print-friendly string."""
        if self.pred is None:
            return "<UNAVAILABLE>"
        confmat = self.matrix(self.gt.numpy(), self.pred.numpy(), self.class_names)
        if self.class_names:
            return "\n" + thelper.utils.stringify_confmat(confmat, self.class_names)
        else:
            return "\n" + str(confmat)

    def render(self):
        """Returns the confusion matrix as a numpy-compatible RGBA image drawn by pyplot."""
        if self.pred is None:
            return None
        confmat = self.matrix(self.gt.numpy(), self.pred.numpy(), self.class_names)
        if self.class_names:
            try:
                fig = thelper.utils.draw_confmat(confmat, self.class_names, normalize=self.draw_normalized)
                array = thelper.utils.fig2array(fig)
                return array
            except AttributeError:
                logger.warning("failed to render confusion matrix figure (caught exception)")
                # return None if rendering fails (probably due to matplotlib on displayless server)
                return None
        else:
            raise NotImplementedError

    def reset(self):
        """Toggles a reset of the metric's internal state, emptying queues."""
        self.pred = None
        self.gt = None

    def goal(self):
        """Returns ``None``, as this class should not be used to directly monitor the training progress."""
        return None


class ROCCurve(Metric):
    """Receiver operating characteristic curve computation interface.

    This class provides an interface to ``sklearn.metrics.roc_curve`` and ``sklearn.metrics.roc_auc_score``
    that can produce various types of ROC-related information including the area under the curve (AUC), the
    false positive and negative rates for various operating points, the ROC curve itself as an image (also
    compatible with tensorboardX), and CSV files containing the metadata of badly predicted samples.

    By default, evaluating this metric returns a print-friendly string containing the AUC score. If a target
    operating point is set, it will instead return the false positive/negative prediction rate of the model
    at that point (also as a print-friendly string). Since this evaluation result is not a scalar, this metric
    cannot be directly used to monitor the progression of a model during a training session, and thus returns
    ``None`` in :func:`thelper.optim.metrics.ROCCurve.goal`.

    Usage examples inside a session configuration file::

        # ...
        # lists all metrics to instantiate as a dictionary
        "metrics": {
            # ...
            # this is the name of the first example; it will output the AUC of the "reject" class
            "roc_reject_auc": {
                # this type is used to instantiate the ROC metric
                "type": "thelper.optim.metrics.ROCCurve",
                # these parameters are passed to the constructor
                "params": {
                    # the name of the class to evaluate
                    "target_name": "reject"
                }
            },
            # this is the name of the second example; it will output the FPR at TPR=0.99
            "roc_reject_0.99tpr": {
                # this type is used to instantiate the ROC metric
                "type": "thelper.optim.metrics.ROCCurve",
                # these parameters are passed to the constructor
                "params": {
                    # the name of the class to evaluate
                    "target_name": "reject",
                    # the target true positive rate (TPR) operating point
                    "target_tpr": 0.99
                }
            },
            # ...
        }
        # ...

    Attributes:
        target_inv: used to target all classes except the named one(s); experimental!
        target_name: name of targeted class to generate the roc curve/auc information for.
        target_tpr: target operating point in terms of true positive rate (provided in constructor).
        target_fpr: target operating point in terms of false positive rate (provided in constructor).
        target_idx: index of the targeted class, mapped from target_name using the class_names list.
        class_names: holds the list of class label names provided by the dataset parser. If it is not
            provided when the constructor is called, it will be set by the trainer at runtime.
        force_softmax: specifies whether a softmax operation should be applied to the prediction scores
            obtained from the trainer.
        log_params: dictionary of extra parameters used to control the logging of bad predictions.
        log_fpr_threshold: used to deduce the fpr operating threshold to use for logging bad predictions.
        log_tpr_threshold: used to deduce the tpr operating threshold to use for logging bad predictions.
        log_meta_keys: list of metadata fields to copy in the log for each bad prediction.
        curve: roc curve generator function, called at evaluation time to generate the output string.
        auc: auc score generator function, called at evaluation time to generate the output string.
        score: queue used to store prediction score values for window-based averaging.
        true: queue used to store groundtruth label values for window-based averaging.
        meta: dictionary of metadata queues used for logging.
    """

    def __init__(self, target_name, target_tpr=None, target_fpr=None, class_names=None,
                 force_softmax=True, log_params=None, sample_weight=None, drop_intermediate=True):
        """Receives the target class/operating point info, log parameters, and roc computation arguments.

        Args:
            target_name: name of targeted class to generate the roc curve/auc information for.
            target_tpr: target operating point in terms of true positive rate (provided in constructor).
            target_fpr: target operating point in terms of false positive rate (provided in constructor).
            class_names: holds the list of class label names provided by the dataset parser. If it is not
                provided when the constructor is called, it will be set by the trainer at runtime.
            force_softmax: specifies whether a softmax operation should be applied to the prediction scores
                obtained from the trainer.
            log_params: dictionary of extra parameters used to control the logging of bad predictions.
            sample_weight: passed to ``sklearn.metrics.roc_curve`` and ``sklearn.metrics.roc_auc_score``.
            drop_intermediate: passed to ``sklearn.metrics.roc_curve``.
        """
        if target_name is None:
            raise AssertionError("must provide a target (class) name for ROC metric")
        self.target_inv = False
        if isinstance(target_name, str) and target_name[0] == "!":
            self.target_inv = True
            self.target_name = target_name.split("!", 1)[1]
        else:
            self.target_name = target_name
        self.target_tpr, self.target_fpr = None, None
        if target_tpr is not None and target_fpr is not None:
            raise AssertionError("must specify only one of target_fpr and target_tpr, not both")
        if target_tpr is not None or target_fpr is not None:
            target_xpr = target_tpr if target_tpr is not None else target_fpr
            if not isinstance(target_xpr, float):
                raise AssertionError("expected float type for target operating point")
            if target_xpr < 0 or target_xpr > 1:
                raise AssertionError("invalid target operation point value (must be in [0,1])")
            if target_tpr is not None:
                self.target_tpr = target_tpr
            if target_fpr is not None:
                self.target_fpr = target_fpr
        self.target_idx = None
        self.class_names = None
        if class_names is not None:
            self.set_class_names(class_names)
        self.force_softmax = force_softmax
        self.log_params = log_params
        if log_params is not None:
            if not isinstance(log_params, dict):
                raise AssertionError("unexpected log params type (expected dict)")
            if "fpr_threshold" not in log_params and "tpr_threshold" not in log_params:
                raise AssertionError("missing log 'fpr_threshold' or 'tpr_threshold' field for logging in params")
            if "fpr_threshold" in log_params and "tpr_threshold" in log_params:
                raise AssertionError("must specify only 'fpr_threshold' or 'tpr_threshold' field for logging in params")
            if "fpr_threshold" in log_params:
                self.log_fpr_threshold = float(log_params["fpr_threshold"])
                if self.log_fpr_threshold < 0 or self.log_fpr_threshold > 1:
                    raise AssertionError("bad log fpr threshold (should be in [0,1]")
                self.log_tpr_threshold = None
            elif "tpr_threshold" in log_params:
                self.log_tpr_threshold = float(log_params["tpr_threshold"])
                if self.log_tpr_threshold < 0 or self.log_tpr_threshold > 1:
                    raise AssertionError("bad log tpr threshold (should be in [0,1]")
                self.log_fpr_threshold = None
            if "meta_keys" not in log_params:
                raise AssertionError("missing log 'meta_keys' field for logging in params")
            self.log_meta_keys = log_params["meta_keys"]
            if not isinstance(self.log_meta_keys, list):
                raise AssertionError("unexpected log meta keys params type (expected list)")

        def gen_curve(y_true, y_score, _target_idx, _target_inv, _sample_weight=sample_weight, _drop_intermediate=drop_intermediate):
            if _target_idx is None:
                raise AssertionError("missing positive target idx at run time")
            _y_true, _y_score = [], []
            for sample_idx, label_idx in enumerate(y_true):
                _y_true.append(label_idx != _target_idx if _target_inv else label_idx == _target_idx)
                _y_score.append(1 - y_score[sample_idx, _target_idx] if _target_inv else y_score[sample_idx, _target_idx])
            res = sklearn.metrics.roc_curve(_y_true, _y_score, sample_weight=_sample_weight, drop_intermediate=_drop_intermediate)
            return res

        def gen_auc(y_true, y_score, _target_idx, _target_inv, _sample_weight=sample_weight):
            if _target_idx is None:
                raise AssertionError("missing positive target idx at run time")
            _y_true, _y_score = [], []
            for sample_idx, label_idx in enumerate(y_true):
                _y_true.append(label_idx != _target_idx if _target_inv else label_idx == _target_idx)
                _y_score.append(1 - y_score[sample_idx, _target_idx] if _target_inv else y_score[sample_idx, _target_idx])
            res = sklearn.metrics.roc_auc_score(_y_true, _y_score, sample_weight=_sample_weight)
            return res

        self.curve = gen_curve
        self.auc = gen_auc
        self.score = None
        self.true = None
        self.meta = None  # needed if outputting tbx txt

    def set_class_names(self, class_names):
        """Sets the class label names that must be predicted by the model.

        This allows the target class name to be mapped to a target class index.

        The current implementation of :class:`thelper.train.trainers.Trainer` will automatically
        call this function at runtime if it is available, and provide the dataset's classes as a
        list of strings.
        """
        if not isinstance(class_names, list):
            raise AssertionError("expected list for class names")
        if len(class_names) < 2:
            raise AssertionError("not enough classes in provided class list")
        if self.target_name not in class_names:
            raise AssertionError("could not find target name '%s' in class names list" % str(self.target_name))
        self.target_idx = class_names.index(self.target_name)
        self.class_names = class_names

    def accumulate(self, pred, gt, meta=None):
        """Receives the latest prediction scores and groundtruth label indices from the trainer.

        Args:
            pred: class prediction scores forwarded by the trainer.
            gt: groundtruth labels forwarded by the trainer (can be ``None`` if unavailable).
            meta: metadata tensors forwarded by the trainer (used for logging, if activated).
        """
        if gt is None:
            return  # only accumulating results when groundtruth available
        if self.force_softmax:
            with torch.no_grad():
                pred = torch.nn.functional.softmax(pred, dim=1)
        if self.score is None:
            self.score = pred.clone()
            self.true = gt.clone()
        else:
            self.score = torch.cat((self.score, pred), 0)
            self.true = torch.cat((self.true, gt), 0)
        if self.log_params:  # do not log meta parameters unless requested
            if meta is None:
                raise AssertionError("sample metadata is required, logging is activated")
            _meta = {key: meta[key] if key in meta else None for key in self.log_meta_keys}
            if self.meta is None:
                self.meta = copy.deepcopy(_meta)
            else:
                for key in self.log_meta_keys:
                    if isinstance(_meta[key], list):
                        self.meta[key] += _meta[key]
                    elif isinstance(_meta[key], torch.Tensor):
                        self.meta[key] = torch.cat((self.meta[key], _meta[key]), 0)
                    elif not (_meta[key] is None and self.meta[key] is None):
                        raise AssertionError("missing impl for meta concat w/ type '%s'" % str(type(_meta[key])))

    def eval(self):
        """Returns a print-friendly string containing the evaluation result (AUC/TPR/FPR).

        If no target operating point is set, the returned string contains the AUC for the target class. If a
        target TPR is set, the returned string contains the FPR for that operating point. If a target FPR is set,
        the returned string contains the TPR for that operating point.
        """
        if self.score is None:
            return "<UNAVAILABLE>"
        # if we did not specify a target operating point in terms of true/false positive rate, return AUC
        if self.target_tpr is None and self.target_fpr is None:
            return "AUC = %.5f" % self.auc(self.true.numpy(), self.score.numpy(), self.target_idx, self.target_inv)
        # otherwise, find the opposite rate at the requested target operating point
        _fpr, _tpr, _thrs = self.curve(self.true.numpy(), self.score.numpy(), self.target_idx, self.target_inv, _drop_intermediate=False)
        for fpr, tpr, thrs in zip(_fpr, _tpr, _thrs):
            if self.target_tpr is not None and tpr >= self.target_tpr:
                return "for target tpr = %.5f, fpr = %.5f at threshold = %f" % (self.target_tpr, fpr, thrs)
            elif self.target_fpr is not None and fpr >= self.target_fpr:
                return "for target fpr = %.5f, tpr = %.5f at threshold = %f" % (self.target_fpr, tpr, thrs)
        # if we did not find a proper rate match above, return worse possible value
        if self.target_tpr is not None:
            return "for target tpr = %.5f, fpr = 1.0 at threshold = min" % self.target_tpr
        elif self.target_fpr is not None:
            return "for target fpr = %.5f, tpr = 0.0 at threshold = max" % self.target_fpr

    def render(self):
        """Returns the ROC curve as a numpy-compatible RGBA image drawn by pyplot."""
        if self.score is None:
            return None
        fpr, tpr, t = self.curve(self.true.numpy(), self.score.numpy(), self.target_idx, self.target_inv)
        try:
            fig = thelper.utils.draw_roc_curve(fpr, tpr)
            array = thelper.utils.fig2array(fig)
            return array
        except AttributeError:
            logger.warning("failed to render confusion matrix figure (caught exception)")
            # return None if rendering fails (probably due to matplotlib on displayless server)
            return None

    def print(self):
        """Returns the logged metadata of badly predicted samples if logging is activated, and ``None`` otherwise.

        The returned object is a print-friendly CSV string that can be consumed directly by tensorboardX. Note
        that this string might be very long if the dataset is large (i.e. it will contain one line per sample).
        """
        if self.log_params is None:
            return None  # do not generate log text unless requested
        if self.meta is None or not self.meta or self.score is None:
            return None
        if self.class_names is None or not self.class_names:
            raise AssertionError("missing class list for logging, current impl only supports named outputs")
        _fpr, _tpr, _t = self.curve(self.true.numpy(), self.score.numpy(), self.target_idx, self.target_inv, _drop_intermediate=False)
        threshold = None
        for fpr, tpr, t in zip(_fpr, _tpr, _t):
            if self.log_fpr_threshold is not None and self.log_fpr_threshold <= fpr:
                threshold = t
                break
            elif self.log_tpr_threshold is not None and self.log_tpr_threshold <= tpr:
                threshold = t
                break
        if threshold is None:
            raise AssertionError("bad fpr/tpr threshold, could not find cutoff for pred scores")
        res = "sample_idx,gt_label_idx,gt_label_name,gt_label_score,pred_label_idx,pred_label_name,pred_label_score"
        for key in self.log_meta_keys:
            res += "," + str(key)
        res += "\n"
        for sample_idx in range(self.true.numel()):
            gt_label_idx = self.true[sample_idx].item()
            scores = self.score[sample_idx, :].tolist()
            gt_label_score = scores[gt_label_idx]
            if (self.target_inv and gt_label_idx != self.target_idx and (1 - gt_label_score) <= threshold) or \
               (not self.target_inv and gt_label_idx == self.target_idx and gt_label_score <= threshold):
                pred_label_score = max(scores)
                pred_label_idx = scores.index(pred_label_score)
                res += "{},{},{},{:2.4f},{},{},{:2.4f}".format(
                    sample_idx,
                    gt_label_idx,
                    self.class_names[gt_label_idx],
                    gt_label_score,
                    pred_label_idx,
                    self.class_names[pred_label_idx],
                    pred_label_score,
                )
                for key in self.log_meta_keys:
                    val = None
                    if key in self.meta and self.meta[key] is not None:
                        val = self.meta[key][sample_idx]
                    if isinstance(val, torch.Tensor) and val.numel() == 1:
                        res += "," + str(val.item())
                    else:
                        res += "," + str(val)
                res += "\n"
        return res

    def reset(self):
        """Toggles a reset of the metric's internal state, emptying queues."""
        self.score = None
        self.true = None
        self.meta = None

    def goal(self):
        """Returns ``None``, as this class should not be used to directly monitor the training progress."""
        return None


class ClassifLogger(Metric):
    """Classification output logger.

    This class provides a simple logging interface for accumulating and saving the predictions of a classifier.
    Note that since the evaluation result is always ``None``, this metric cannot be used to directly monitor
    training progression, and thus also returns ``None`` in :func:`thelper.optim.metrics.ClassifLogger.goal`.

    It also optionally offers tensorboardX-compatible output images that can be saved locally or posted to
    tensorboard for browser-based visualization.

    Usage examples inside a session configuration file::

        # ...
        # lists all metrics to instantiate as a dictionary
        "metrics": {
            # ...
            # this is the name of the example metric; it is used for lookup/printing only
            "logger": {
                # this type is used to instantiate the confusion matrix report metric
                "type": "thelper.optim.metrics.ClassifLogger",
                "params": {
                    # log the three 'best' predictions for each sample
                    "top_k": 3,
                    # keep updating a set of 10 samples for visualization via tensorboardX
                    "viz_count": 10
                }
            },
            # ...
        }
        # ...

    Attributes:
        top_k: number of 'best' predictions to keep for each sample (along with the gt label).
        class_names: holds the list of class label names provided by the dataset parser. If it is not
            provided when the constructor is called, it will be set by the trainer at runtime.
        viz_count: number of tensorboardX images to generate and update at each epoch.
        meta_keys: list of metadata fields to copy in the log for each prediction.
        force_softmax: specifies whether a softmax operation should be applied to the prediction scores
            obtained from the trainer.
        score: queue used to store prediction score values for logging.
        true: queue used to store groundtruth label values for logging.
        meta: dictionary of metadata queues used for logging.
    """

    def __init__(self, top_k=1, class_names=None, viz_count=0, meta_keys=None, force_softmax=True):
        """Receives the logging parameters & the optional class label names used to decorate the log.

        Args:
            top_k: number of 'best' predictions to keep for each sample (along with the gt label).
            class_names: holds the list of class label names provided by the dataset parser. If it is not
                provided when the constructor is called, it will be set by the trainer at runtime.
            viz_count: number of tensorboardX images to generate and update at each epoch.
            meta_keys: list of metadata fields to copy in the log for each prediction.
            force_softmax: specifies whether a softmax operation should be applied to the prediction scores
                obtained from the trainer.
        """
        if not isinstance(top_k, int) or top_k <= 0:
            raise AssertionError("invalid top-k value")
        self.top_k = top_k
        self.class_names = None
        if class_names is not None:
            self.set_class_names(class_names)
        if not isinstance(viz_count, int) or top_k < 0:
            raise AssertionError("invalid viz_count value")
        self.viz_count = viz_count
        if meta_keys is not None and not isinstance(meta_keys, list):
            raise AssertionError("unexpected log meta keys params type (expected list)")
        self.meta_keys = meta_keys
        self.force_softmax = force_softmax
        self.score = None
        self.true = None
        self.meta = None

    def set_class_names(self, class_names):
        """Sets the class label names that must be predicted by the model.

        This allows the target class name to be mapped to a target class index.

        The current implementation of :class:`thelper.train.trainers.Trainer` will automatically
        call this function at runtime if it is available, and provide the dataset's classes as a
        list of strings.
        """
        if not isinstance(class_names, list):
            raise AssertionError("expected list for class names")
        if len(class_names) < 2:
            raise AssertionError("not enough classes in provided class list")
        self.class_names = class_names

    def accumulate(self, pred, gt, meta=None):
        """Receives the latest prediction scores and groundtruth label indices from the trainer.

        Args:
            pred: class prediction scores forwarded by the trainer.
            gt: groundtruth labels forwarded by the trainer (can be ``None`` if unavailable).
            meta: metadata tensors forwarded by the trainer (used for logging).
        """
        if self.force_softmax:
            with torch.no_grad():
                pred = torch.nn.functional.softmax(pred, dim=1)
        if gt is None:
            # if gt is missing, set labels as -1; will be reinterpreted when creating log
            gt = torch.from_numpy(np.full((pred.shape[0], 1), -1, dtype=np.int64))
        if self.score is None:
            self.score = pred.clone()
            self.true = gt.clone()
        else:
            self.score = torch.cat((self.score, pred), 0)
            self.true = torch.cat((self.true, gt), 0)
        if self.meta_keys is not None and self.meta_keys:
            if meta is None:
                raise AssertionError("sample metadata is required, logging is activated")
            _meta = {key: meta[key] if key in meta else None for key in self.meta_keys}
            if self.meta is None:
                self.meta = copy.deepcopy(_meta)
            else:
                for key in self.meta_keys:
                    if isinstance(_meta[key], list):
                        self.meta[key] += _meta[key]
                    elif isinstance(_meta[key], torch.Tensor):
                        self.meta[key] = torch.cat((self.meta[key], _meta[key]), 0)
                    else:
                        raise AssertionError("missing impl for meta concat w/ type '%s'" % str(type(_meta[key])))

    def eval(self):
        """Returns ``None``, as this class only produces log files in the session directory."""
        return None

    def render(self):
        """Returns an image of predicted outputs as a numpy-compatible RGBA image drawn by pyplot."""
        if self.viz_count == 0:
            return None
        raise NotImplementedError  # TODO

    def print(self):
        """Returns the logged metadata of predicted samples.

        The returned object is a print-friendly CSV string that can be consumed directly by tensorboardX. Note
        that this string might be very long if the dataset is large (i.e. it will contain one line per sample).
        """
        if self.score is None:
            return None
        if self.class_names is None or not self.class_names:
            raise AssertionError("missing class list for logging, current impl only supports named outputs")
        res = "sample_idx,gt_label_idx,gt_label_name,gt_label_score"
        for k in range(self.top_k):
            res += ",pred_label_idx_%d,pred_label_name_%d,pred_label_score_%d" % (k, k, k)
        if self.meta_keys is not None and self.meta_keys:
            for key in self.meta_keys:
                res += "," + str(key)
        res += "\n"
        for sample_idx in range(self.true.numel()):
            gt_label_idx = self.true[sample_idx].item()
            scores = np.asarray(self.score[sample_idx, :].tolist())
            sorted_score_idxs = np.argsort(scores)[::-1]
            sorted_scores = scores[sorted_score_idxs]
            res += "{},{},{},{:2.4f}".format(
                sample_idx,
                gt_label_idx,
                self.class_names[gt_label_idx] if gt_label_idx >= 0 else "n/a",
                scores[gt_label_idx] if gt_label_idx >= 0 else 0
            )
            for k in range(self.top_k):
                res += ",{},{},{:2.4f}".format(
                    sorted_score_idxs[k],
                    self.class_names[sorted_score_idxs[k]],
                    sorted_scores[k]
                )
            if self.meta_keys is not None and self.meta_keys:
                for key in self.meta_keys:
                    val = None
                    if key in self.meta and self.meta[key] is not None:
                        val = self.meta[key][sample_idx]
                    if isinstance(val, torch.Tensor) and val.numel() == 1:
                        res += "," + str(val.item())
                    else:
                        res += "," + str(val)
            res += "\n"
        return res

    def reset(self):
        """Toggles a reset of the metric's internal state, emptying queues."""
        self.score = None
        self.true = None
        self.meta = None

    def goal(self):
        """Returns ``None``, as this class should not be used to directly monitor the training progress."""
        return None


class RawPredictions(Metric):
    """Raw predictions storage.

    This class provides a simple interface for accumulating and saving the raw predictions of a classifier.
    Note that since the evaluation result is always ``None``, this metric cannot be used to directly monitor
    training progression, and thus also returns ``None`` in :func:`thelper.optim.metrics.ClassifLogger.goal`.

    It also optionally offers a callback functionality on each accumulated prediction to execute additional
    operations (ex: call a function to update external processes from ``thelper`` package).

    Usage examples inside a session configuration file::

        # ...
        # lists all metrics to instantiate as a dictionary
        "metrics": {
            # ...
            # this is the name of the example metric; it is used for lookup/printing only
            "predictions": {
                # this type is used to instantiate the confusion matrix report metric
                "type": "thelper.optim.metrics.RawPredictions",
                "params": [
                    # call 'my_function' located within 'external_module' after each added prediction
                    {"name": "callback", "value": "external_module.my_function"},
                ]
            },
            # ...
        }
        # ...

    Attributes:
        callback: callable to be executed after each accumulated prediction
    """

    def __init__(self, callback=None):
        # type: (Optional[Callable]) -> None
        super().__init__()
        self.predictions = list()   # type: List[Dict[AnyStr, Any]]
        if callback is not None and not callable(callback):
            raise TypeError("Callback is not callable, got {!s}.".format(type(callback)))
        self.callback = callback or (lambda *args, **kwargs: None)  # do nothing if None to simplify calls

    @staticmethod
    def _to_py(element):
        if isinstance(element, torch.Tensor):
            return element.tolist()
        return element

    def accumulate(self, pred, gt, meta=None):
        """Receives the latest prediction and groundtruth tensors (each batch) from the session.

        Args:
            pred: model prediction tensor forwarded by the trainer for a given sample batch.
            gt: groundtruth tensor forwarded by the trainer (can be ``None`` if unavailable).
            meta: metadata tensors forwarded by the trainer.
        """

        # convert batch tensor to per-sample class predictions
        samples_predictions = [{
            "predictions": self._to_py(p),
        } for p in pred]

        # convert ground truths to per-sample labels
        for i, label in enumerate(self._to_py(gt)):
            samples_predictions[i]["gt"] = label

        # transfer meta information to corresponding samples according to available details
        n_samples = len(samples_predictions)
        if isinstance(meta, dict):
            for meta_key in meta:
                if isinstance(meta[meta_key], list):
                    n_meta = len(meta[meta_key])
                    # list of tensors where each index in every tensor corresponds to a sample index
                    if n_meta != n_samples:
                        meta_info = [self._to_py(t) for t in meta[meta_key]]
                        for j, s in enumerate(samples_predictions):
                            # transfer corresponding sample index info to matching tensor prediction,
                            # or transfer whole meta info if index correspondence cannot be established
                            s[meta_key] = [meta_info[i][j] if hasattr(meta_info[i], '__len__') else meta_info[i]
                                           for i in range(len(meta_info))]
                    # list of elements or tensor of samples size
                    elif n_meta == n_samples:
                        meta_info = self._to_py(meta[meta_key])
                        for i, s in enumerate(samples_predictions):
                            s[meta_key] = meta_info[i]
                else:
                    # each sample gets the same info
                    for s in samples_predictions:
                        s[meta_key] = meta[meta_key]

        self.predictions.extend(samples_predictions)
        self.callback()

    def reset(self):
        self.__init__(callback=self.callback)

    def eval(self):
        """Returns ``None``, as this class only preserves raw prediction results."""
        return None

    def goal(self):
        """Returns ``None``, as this class should not be used to directly monitor the training progress."""
        return None



"""Transformations wrappers module.

The wrapper classes herein are used to either support inline operations on odd sample types (e.g. lists
of images) or for external libraries (e.g. Augmentor).
"""

import logging
import random

import numpy as np
import PIL.Image
import torch

import thelper.utils

logger = logging.getLogger(__name__)


class AlbumentationsWrapper(object):
    """Albumentations pipeline wrapper that allows dictionary unpacking.

    See https://github.com/albu/albumentations for more information.

    Attributes:
        pipeline: the augmentor pipeline instance to apply to images.
        image_key: the key to fetch images from (when dictionaries are passed in).
        bboxes_key: the key to fetch bounding boxes from (when dictionaries are passed in).
        mask_key: the key to fetch masks from (when dictionaries are passed in).
        keypoints_key: the key to fetch keypoints from (when dictionaries are passed in).
        cvt_kpts_to_bboxes: specifies whether keypoints should be converted to bboxes for compatbility.
        linked_fate: specifies whether input list samples should all have the same fate or not.

    .. seealso::
        | :func:`thelper.transforms.utils.load_transforms`
    """

    def __init__(self, transforms, to_tensor=None, bbox_params=None, add_targets=None, image_key="image",
                 bboxes_key="bboxes", mask_key="mask", keypoints_key="keypoints", probability=1.0,
                 cvt_kpts_to_bboxes=False, linked_fate=False):
        """Receives and stores an augmentor pipeline for later use.

        The pipeline itself is instantiated in :func:`thelper.transforms.utils.load_transforms`.
        """
        if bbox_params is None or not bbox_params:
            bbox_params = {"format": "coco"}  # i.e. opencv format (X,Y,W,H)
        if add_targets is None:
            add_targets = {}
        if isinstance(image_key, (list, tuple)):
            if len(image_key) > 1:
                raise AssertionError("current implementation cannot handle more than one input image key per packet")
            image_key = image_key[0]
        self.image_key = image_key
        if isinstance(bboxes_key, (list, tuple)) or isinstance(keypoints_key, (list, tuple)) or isinstance(mask_key, (list, tuple)):
            raise AssertionError("bboxes/keypoints/masks keys should never be passed as lists")
        self.bboxes_key = bboxes_key
        self.mask_key = mask_key
        self.keypoints_key = keypoints_key
        self.cvt_kpts_to_bboxes = cvt_kpts_to_bboxes
        if cvt_kpts_to_bboxes and "format" not in bbox_params or bbox_params["format"] != "coco":
            raise AssertionError("if converting kpts to bboxes, must use coco format")
        self.linked_fate = linked_fate
        import albumentations
        self.pipeline = albumentations.Compose(transforms, to_tensor=to_tensor, bbox_params=bbox_params,
                                               additional_targets=add_targets, p=probability)

    def __call__(self, sample, force_linked_fate=False, op_seed=None):
        """Transforms a (dict) sample, a single image, or a list of images using the augmentor pipeline.

        Args:
            sample: the sample or image(s) to transform (can also contain embedded lists/tuples of images).
            force_linked_fate: override flag for recursive use allowing forced linking of arrays.
            op_seed: seed to set before calling the wrapped operation.

        Returns:
            The transformed image(s), with the same list/tuple formatting as the input.
        """
        # todo: add list unwrapping/interlacing support like in other wrappers?
        params = {}
        if isinstance(sample, dict):
            if self.image_key not in sample:
                raise AssertionError("image is missing from sample (key=%s) but it is mandatory" % self.image_key)
            image = sample[self.image_key]
            if isinstance(image, (list, tuple)):
                raise NotImplementedError
                # impl should use linked_fate and force_linked_fate
            params["image"] = sample[self.image_key]
            if self.keypoints_key in sample and sample[self.keypoints_key] is not None:
                keypoints = sample[self.keypoints_key]
                if self.cvt_kpts_to_bboxes:
                    if self.bboxes_key in sample:
                        raise AssertionError("trying to override bboxes w/ keypoints while bboxes already exist")
                    # fake x,y,w,h,c format (w/ labels)
                    msize = params["image"].shape
                    params["bboxes"] = [[min(max(kp[0], 0), msize[1] - 1),
                                         min(max(kp[1], 0), msize[0] - 1), 1, 1, 0] for kp in keypoints]
                else:
                    params["keypoints"] = keypoints
            if self.bboxes_key in sample and sample[self.bboxes_key] is not None:
                params["bboxes"] = sample[self.bboxes_key]
            else:
                params["bboxes"] = []
            if self.mask_key in sample and sample[self.mask_key] is not None:
                params["mask"] = sample[self.mask_key]
            output = self.pipeline(**params)
            sample[self.image_key] = output["image"]
            if "keypoints" in output:
                sample[self.keypoints_key] = output["keypoints"]
            if "bboxes" in output:
                if self.cvt_kpts_to_bboxes:
                    sample[self.keypoints_key] = [[kp[0], kp[1]] for kp in output["bboxes"]]
                else:
                    sample[self.bboxes_key] = output["bboxes"]
            if "mask" in output:
                sample[self.mask_key] = output["mask"]
            return sample
        elif isinstance(sample, (list, tuple)):
            raise NotImplementedError
            # impl should use linked_fate and force_linked_fate
        else:
            if sample is None:
                return None
            elif not isinstance(sample, np.ndarray):
                raise AssertionError("unexpected input image type")
            params["image"] = sample
        output = self.pipeline(**params)
        return output["image"]

    def __repr__(self):
        """Create a print-friendly representation of inner augmentation stages."""
        return self.__class__.__name__ + (": {{image_key: {}, bboxes_key: {}, mask_key: {}, keypoints_key: {}, "
                                          .format(self.image_key, self.bboxes_key, self.mask_key, self.keypoints_key) +
                                          "cvt_kpts_to_bboxes: {}, linked_fate: {}, pipeline: {}"
                                          .format(self.cvt_kpts_to_bboxes, self.linked_fate, self.pipeline) + "}")

    # noinspection PyMethodMayBeStatic
    def set_seed(self, seed):
        """Sets the internal seed to use for stochastic ops."""
        if self.pipeline.transforms is not None:
            random.random(seed)
            np.random.seed(seed)
            for t in self.pipeline.transforms:
                if hasattr(t, "set_seed") and callable(t.set_seed):
                    t.set_seed(seed)

    def set_epoch(self, epoch=0):
        """Sets the current epoch number in order to change the behavior of some suboperations."""
        if not isinstance(epoch, int) or epoch < 0:
            raise AssertionError("invalid epoch value")
        if self.pipeline.transforms is not None:
            for t in self.pipeline.transforms:
                if hasattr(t, "set_epoch") and callable(t.set_epoch):
                    t.set_epoch(epoch)


class AugmentorWrapper(object):
    """Augmentor pipeline wrapper that allows pickling and multi-threading.

    See https://github.com/mdbloice/Augmentor for more information. This wrapper was last updated to work
    with version 0.2.2 --- more recent versions introduced yet unfixed (as of 2018/08) issues on some platforms.

    All original transforms are supported here. This wrapper also fixes the list output bug for single-image
    samples when using operations individually.

    Attributes:
        pipeline: the augmentor pipeline instance to apply to images.
        target_keys: the sample keys to apply the pipeline to (when dictionaries are passed in).
        linked_fate: specifies whether input list samples should all have the same fate or not.

    .. seealso::
        | :func:`thelper.transforms.utils.load_transforms`
    """

    def __init__(self, pipeline, target_keys=None, linked_fate=True):
        """Receives and stores an augmentor pipeline for later use.

        The pipeline itself is instantiated in :func:`thelper.transforms.utils.load_transforms`.
        """
        self.pipeline = pipeline
        self.target_keys = target_keys
        self.linked_fate = linked_fate

    def __call__(self, sample, force_linked_fate=False, op_seed=None, in_cvts=None):
        """Transforms a (dict) sample, a single image, or a list of images using the augmentor pipeline.

        Args:
            sample: the sample or image(s) to transform (can also contain embedded lists/tuples of images).
            force_linked_fate: override flag for recursive use allowing forced linking of arrays.
            op_seed: seed to set before calling the wrapped operation.
            in_cvts: holds the input conversion flag array (for recursive usage).

        Returns:
            The transformed image(s), with the same list/tuple formatting as the input.
        """
        if isinstance(sample, dict):
            # recursive call for unpacking sample content w/ target keys
            if in_cvts is not None:
                raise AssertionError("top-level call should never provide in_cvts")
            # capture all array-like objects via __getitem__ test (if no keys are provided)
            key_vals = [(k, v) for k, v in sample.items() if (
                (self.target_keys is None and hasattr(v, "__getitem__") and not isinstance(v, str)) or
                (self.target_keys is not None and k in self.target_keys))]
            keys, vals = map(list, zip(*key_vals))
            lengths = [len(v) if isinstance(v, (list, tuple)) else -1 for v in vals]
            if len(lengths) > 0 and all(n == lengths[0] for n in lengths) and lengths[0] > 0:
                # interlace input lists for internal linked fate (if needed; otherwise, it won't change anything)
                vals = [[v[idx] if isinstance(v, (list, tuple)) else
                         v[idx, ...] for v in vals] for idx in range(lengths[0])]
                vals = self(vals, force_linked_fate=force_linked_fate, op_seed=op_seed, in_cvts=in_cvts)
                if not isinstance(vals, list) or len(vals) != lengths[0]:
                    raise AssertionError("messed up something internally")
                out_vals = [[v] for v in vals[0]] if isinstance(vals[0], list) else [[vals[0]]]
                for idx1 in range(1, lengths[0]):
                    for idx2 in range(len(out_vals)):
                        out_vals[idx2].append(vals[idx1][idx2] if isinstance(vals[idx1], list) else vals[idx1])
                vals = out_vals
            else:
                vals = self(vals, force_linked_fate=force_linked_fate, op_seed=op_seed, in_cvts=in_cvts)
            sample = {k: vals[keys.index(k)] if k in keys else sample[k] for k in sample}
            return sample
        out_cvts = in_cvts is not None
        out_list = isinstance(sample, (list, tuple))
        if sample is None or (out_list and not sample):
            return ([], []) if out_cvts else []
        elif not out_list:
            sample = [sample]
        if any([isinstance(v, dict) for v in sample]):
            raise AssertionError("augmentor wrapper cannot handle sample-in-sample (or dict-in-list) inputs")
        skip_unpack = in_cvts is not None and isinstance(in_cvts, bool) and in_cvts
        if self.linked_fate or force_linked_fate:  # process all content with the same operations below
            if not skip_unpack:
                # noinspection PyProtectedMember
                sample, cvts = TransformWrapper._unpack(sample, convert_pil=True)
                if not isinstance(sample, (list, tuple)):
                    sample = [sample]
                    cvts = [cvts]
            else:
                cvts = in_cvts
            if op_seed is None:
                op_seed = np.random.randint(np.iinfo(np.int32).max)
            np.random.seed(op_seed)
            prev_state = np.random.get_state()
            for idx, _ in enumerate(sample):
                if not isinstance(sample[idx], PIL.Image.Image):
                    sample[idx], cvts[idx] = self(sample[idx], force_linked_fate=True,
                                                  op_seed=op_seed, in_cvts=cvts[idx])
                else:
                    np.random.set_state(prev_state)
                    random.seed(np.random.randint(np.iinfo(np.int32).max))
                    for operation in self.pipeline.operations:
                        r = round(np.random.uniform(0, 1), 1)
                        if r <= operation.probability:
                            if sample[idx] is not None:
                                sample[idx] = operation.perform_operation([sample[idx]])[0]
        else:  # each element of the top array will be processed independently below (current seeds are kept)
            cvts = [False] * len(sample)
            for idx, _ in enumerate(sample):
                # noinspection PyProtectedMember
                sample[idx], cvts[idx] = TransformWrapper._unpack(sample[idx], convert_pil=True)
                if not isinstance(sample[idx], PIL.Image.Image):
                    sample[idx], cvts[idx] = self(sample[idx], force_linked_fate=True,
                                                  op_seed=op_seed, in_cvts=cvts[idx])
                else:
                    random.seed(np.random.randint(np.iinfo(np.int32).max))
                    for operation in self.pipeline.operations:
                        r = round(np.random.uniform(0, 1), 1)
                        if r <= operation.probability:
                            if sample[idx] is not None:
                                sample[idx] = operation.perform_operation([sample[idx]])[0]
        # noinspection PyProtectedMember
        sample, cvts = TransformWrapper._pack(sample, cvts, convert_pil=True)
        if len(sample) != len(cvts):
            raise AssertionError("messed up packing/unpacking logic")
        if (skip_unpack or not out_list) and len(sample) == 1:
            sample = sample[0]
            cvts = cvts[0]
        return (sample, cvts) if out_cvts else sample

    def __repr__(self):
        """Create a print-friendly representation of inner augmentation stages."""
        return self.__class__.__name__ + (": {{target_keys: {}, linked_fate: {}, "
                                          .format(self.target_keys, self.linked_fate) +
                                          ", ".join([str(t) for t in self.pipeline.operations]) + "}")

    # noinspection PyMethodMayBeStatic
    def set_seed(self, seed):
        """Sets the internal seed to use for stochastic ops."""
        np.random.seed(seed)

    def set_epoch(self, epoch=0):
        """Sets the current epoch number in order to change the behavior of some suboperations."""
        if not isinstance(epoch, int) or epoch < 0:
            raise AssertionError("invalid epoch value")
        if self.pipeline.operations is not None:
            for op in self.pipeline.operations:
                if hasattr(op, "set_epoch") and callable(op.set_epoch):
                    op.set_epoch(epoch)


class TransformWrapper(object):
    """Transform wrapper that allows operations on samples, lists, tuples, and single elements.

    Can be used to wrap the operations in ``thelper.transforms`` or in ``torchvision.transforms``
    that only accept array-like objects as input. Will optionally force-convert content to PIL images.

    Can also be used to transform a list/tuple of images uniformly based on a shared dice roll, or
    to ensure that each image is transformed independently.

    .. warning::
        Stochastic transforms (e.g. ``torchvision.transforms.RandomCrop``) will always treat each
        image in a list differently. If the same operations are to be applied to all images, you
        should consider using a series non-stochastic operations wrapped inside an instance of
        ``torchvision.transforms.RandomApply``, or simply provide the probability of applying the
        transforms to this wrapper's constructor.

    Attributes:
        operation: the wrapped operation (callable object or class name string to import).
        params: the parameters that are passed to the operation when init'd or called.
        probability: the probability that the wrapped operation will be applied.
        convert_pil: specifies whether images should be converted into PIL format or not.
        target_keys: the sample keys to apply the transform to (when dictionaries are passed in).
        linked_fate: specifies whether images given in a list/tuple should have the same fate or not.
    """

    def __init__(self, operation, params=None, probability=1, convert_pil=False, target_keys=None, linked_fate=True):
        """Receives and stores a torchvision transform operation for later use.

        If the operation is given as a string, it is assumed to be a class name and it will
        be imported. The parameters (if any) will then be given to the constructor of that
        class. Otherwise, the operation is assumed to be a callable object, and its parameters
        (if any) will be provided at call-time.

        Args:
            operation: the wrapped operation (callable object or class name string to import).
            params: the parameters that are passed to the operation when init'd or called.
            probability: the probability that the wrapped operation will be applied.
            convert_pil: specifies whether images should be forced into PIL format or not.
            target_keys: the sample keys to apply the pipeline to (when dictionaries are passed in).
            linked_fate: specifies whether images given in a list/tuple should have the same fate or not.
        """
        if params is not None and not isinstance(params, dict):
            raise AssertionError("expected params to be passed in as a dictionary")
        if isinstance(operation, str):
            operation_type = thelper.utils.import_class(operation)
            self.operation = operation_type(**params) if params is not None else operation_type()
            self.params = {}
        else:
            self.operation = operation
            self.params = params if params is not None else {}
        if probability < 0 or probability > 1:
            raise AssertionError("invalid probability value (range is [0,1]")
        self.probability = probability
        self.convert_pil = convert_pil
        self.target_keys = target_keys
        self.linked_fate = linked_fate

    @staticmethod
    def _unpack(sample, force_flatten=False, convert_pil=False):
        if isinstance(sample, (list, tuple)):
            if len(sample) > 1:
                if not force_flatten:
                    return sample, [False] * len(sample)
                flat_samples = []
                cvts = []
                for s in sample:
                    out, cvt = TransformWrapper._unpack(s, force_flatten=force_flatten)
                    if isinstance(cvt, (list, tuple)):
                        if not isinstance(out, (list, tuple)):
                            raise AssertionError("unexpected out/cvt types")
                        flat_samples += list(out)
                        cvts += list(cvt)
                    else:
                        flat_samples.append(out)
                        cvts.append(cvt)
                return flat_samples, cvts
            else:
                sample = sample[0]
        if convert_pil:
            if isinstance(sample, torch.Tensor):
                sample = sample.numpy()
            if isinstance(sample, np.ndarray) and sample.ndim > 2 and \
                    sample.shape[-1] > 1 and (sample.dtype != np.uint8):
                # PIL images cannot handle multi-channel non-byte arrays; we handle these manually
                flat_samples = []
                for c in range(sample.shape[-1]):
                    flat_samples.append(PIL.Image.fromarray(sample[..., c]))
                return flat_samples, True  # this is the only case where an array can be paired with a single cvt flag
            else:
                out = PIL.Image.fromarray(np.squeeze(sample))
                return out, True
        return sample, False

    @staticmethod
    def _pack(samples, cvts, convert_pil=False):
        if not isinstance(samples, (list, tuple)) or not isinstance(cvts, (list, tuple)) or len(samples) != len(cvts):
            if not convert_pil or not isinstance(cvts, bool) or not cvts:
                raise AssertionError("unexpected cvts len w/ pil conversion (bad logic somewhere)")
            if not all([isinstance(s, PIL.Image.Image) for s in samples]):
                raise AssertionError("unexpected packed list sample types")
            samples = [np.asarray(s) for s in samples]
            if not all([s.ndim == 2 for s in samples]):
                raise AssertionError("unexpected packed list sample depths")
            samples = [np.expand_dims(s, axis=2) for s in samples]
            return [np.concatenate(samples, axis=2)], [False]
        for idx, cvt in enumerate(cvts):
            if not isinstance(cvt, (list, tuple)):
                if not isinstance(cvt, bool):
                    raise AssertionError("unexpected cvt type")
                if cvt:
                    if isinstance(samples[idx], (list, tuple)):
                        raise AssertionError("unexpected packed sample type")
                    samples[idx] = np.asarray(samples[idx])
                    cvts[idx] = False
        return samples, cvts

    def __call__(self, sample, force_linked_fate=False, op_seed=None, in_cvts=None):
        """Transforms a (dict) sample, a single image, or a list of images using a wrapped operation.

        Args:
            sample: the sample or image(s) to transform (can also contain embedded lists/tuples of images).
            force_linked_fate: override flag for recursive use allowing forced linking of arrays.
            op_seed: seed to set before calling the wrapped operation.
            in_cvts: holds the input conversion flag array (for recursive usage).

        Returns:
            The transformed image(s), with the same list/tuple formatting as the input.
        """
        if isinstance(sample, dict):
            # recursive call for unpacking sample content w/ target keys
            if in_cvts is not None:
                raise AssertionError("top-level call should never provide in_cvts")
            # capture all array-like objects via __getitem__ test (if no keys are provided)
            key_vals = [(k, v) for k, v in sample.items() if (
                (self.target_keys is None and hasattr(v, "__getitem__") and not isinstance(v, str)) or
                (self.target_keys is not None and k in self.target_keys))]
            keys, vals = map(list, zip(*key_vals))
            lengths = [len(v) if isinstance(v, (list, tuple)) else -1 for v in vals]
            if len(lengths) > 0 and all(n == lengths[0] for n in lengths) and lengths[0] > 0:
                # interlace input lists for internal linked fate (if needed; otherwise, it won't change anything)
                vals = [[v[idx] if isinstance(v, (list, tuple)) else
                         v[idx, ...] for v in vals] for idx in range(lengths[0])]
                vals = self(vals, force_linked_fate=force_linked_fate, op_seed=op_seed, in_cvts=in_cvts)
                if not isinstance(vals, list) or len(vals) != lengths[0]:
                    raise AssertionError("messed up something internally")
                out_vals = [[v] for v in vals[0]] if isinstance(vals[0], list) else [[vals[0]]]
                for idx1 in range(1, lengths[0]):
                    for idx2 in range(len(out_vals)):
                        out_vals[idx2].append(vals[idx1][idx2] if isinstance(vals[idx1], list) else vals[idx1])
                vals = out_vals
            else:
                vals = self(vals, force_linked_fate=force_linked_fate, op_seed=op_seed, in_cvts=in_cvts)
            sample = {k: vals[keys.index(k)] if k in keys else sample[k] for k in sample}
            return sample
        out_cvts = in_cvts is not None
        out_list = isinstance(sample, (list, tuple))
        if sample is None or (out_list and not sample):
            return ([], []) if out_cvts else []
        elif not out_list:
            sample = [sample]
        if any([isinstance(v, dict) for v in sample]):
            raise AssertionError("sample transform wrapper cannot handle sample-in-sample (or dict-in-list) inputs")
        skip_unpack = in_cvts is not None and isinstance(in_cvts, bool) and in_cvts
        if self.linked_fate or force_linked_fate:  # process all content with the same operations below
            if not skip_unpack:
                sample, cvts = self._unpack(sample, convert_pil=self.convert_pil)
                if not isinstance(sample, (list, tuple)):
                    sample = [sample]
                    cvts = [cvts]
            else:
                cvts = in_cvts
            if self.probability >= 1 or round(np.random.uniform(0, 1), 1) <= self.probability:
                if op_seed is None:
                    op_seed = np.random.randint(np.iinfo(np.int32).max)
                for idx, _ in enumerate(sample):
                    if isinstance(sample[idx], (list, tuple)):
                        sample[idx], cvts[idx] = self(sample[idx], force_linked_fate=True,
                                                      op_seed=op_seed, in_cvts=cvts[idx])
                    else:
                        if hasattr(self.operation, "set_seed") and callable(self.operation.set_seed):
                            self.operation.set_seed(op_seed)
                        # watch out: if operation is stochastic and we cannot seed above, then there is no
                        # guarantee that the content will truly have a 'linked fate' (this might cause issues!)
                        if sample[idx] is not None:
                            sample[idx] = self.operation(sample[idx], **self.params)
        else:  # each element of the top array will be processed independently below (current seeds are kept)
            cvts = [False] * len(sample)
            for idx, _ in enumerate(sample):
                sample[idx], cvts[idx] = self._unpack(sample[idx], convert_pil=self.convert_pil)
                if self.probability >= 1 or round(np.random.uniform(0, 1), 1) <= self.probability:
                    if isinstance(sample[idx], (list, tuple)):
                        # we will now force fate linkage for all sub-elements of this array
                        sample[idx], cvts[idx] = self(sample[idx], force_linked_fate=True,
                                                      op_seed=op_seed, in_cvts=cvts[idx])
                    else:
                        if sample[idx] is not None:
                            sample[idx] = self.operation(sample[idx], **self.params)
        sample, cvts = TransformWrapper._pack(sample, cvts, convert_pil=self.convert_pil)
        if len(sample) != len(cvts):
            raise AssertionError("messed up packing/unpacking logic")
        if (skip_unpack or not out_list) and len(sample) == 1:
            sample = sample[0]
            cvts = cvts[0]
        return (sample, cvts) if out_cvts else sample

    def __repr__(self):
        """Create a print-friendly representation of inner augmentation stages."""
        return self.__class__.__name__ + ": {{target_keys: {}, linked_fate: {}, probability: {}. operation: {}".format(
            self.target_keys, self.linked_fate, self.probability, str(self.operation))

    # noinspection PyMethodMayBeStatic
    def set_seed(self, seed):
        """Sets the internal seed to use for stochastic ops."""
        np.random.seed(seed)

    def set_epoch(self, epoch=0):
        """Sets the current epoch number in order to change the behavior of some suboperations."""
        if not isinstance(epoch, int) or epoch < 0:
            raise AssertionError("invalid epoch value")
        if self.operation is not None:
            if hasattr(self.operation, "set_epoch") and callable(self.operation.set_epoch):
                self.operation.set_epoch(epoch)



"""Dataset parsing/loading package.

This package contains classes and functions whose role is to fetch the data required to train, validate,
and test a model. The :func:`thelper.data.utils.create_loaders` function contained herein is responsible for
preparing the task and data loaders for a training session. This package also contains the base interfaces
for dataset parsers.
"""

import logging

import thelper.data.loaders  # noqa: F401
import thelper.data.parsers  # noqa: F401
import thelper.data.pascalvoc  # noqa: F401
import thelper.data.samplers  # noqa: F401
import thelper.data.utils  # noqa: F401
from thelper.data.loaders import DataLoader  # noqa: F401
from thelper.data.parsers import ClassificationDataset  # noqa: F401
from thelper.data.parsers import Dataset  # noqa: F401
from thelper.data.parsers import ExternalDataset  # noqa: F401
from thelper.data.parsers import ImageDataset  # noqa: F401
from thelper.data.parsers import ImageFolderDataset  # noqa: F401
from thelper.data.pascalvoc import PASCALVOC  # noqa: F401
from thelper.data.samplers import SubsetRandomSampler  # noqa: F401
from thelper.data.samplers import SubsetSequentialSampler  # noqa: F401
from thelper.data.samplers import WeightedSubsetRandomSampler  # noqa: F401
from thelper.data.utils import create_loaders  # noqa: F401
from thelper.data.utils import create_parsers  # noqa: F401
from thelper.data.utils import get_class_weights  # noqa: F401

logger = logging.getLogger("thelper.data")



"""Task definition package.

This package contains classes and functions whose role is to define the input/output formats and operations
expected from a trained model. This essentially defines the 'goal' of the model, and is used to specialize
and automate its training.
"""

import logging

from thelper.tasks.classif import Classification  # noqa: F401
from thelper.tasks.regr import Regression  # noqa: F401
from thelper.tasks.segm import Segmentation  # noqa: F401
from thelper.tasks.utils import Task  # noqa: F401
from thelper.tasks.utils import create_global_task  # noqa: F401
from thelper.tasks.utils import create_task  # noqa: F401

logger = logging.getLogger("thelper.tasks")



"""
Command-line module, for use with a ``__main__`` entrypoint.

This module contains the primary functions used to create or resume a training session, to start a
visualization session, or to start an annotation session. The basic argument that needs to be provided
by the user to create any kind of session is a configuration dictionary. For sessions that produce
outputs, the path to a directory where to save the data is also needed.
"""

import argparse
import json
import logging
import os
import sys

import torch

import thelper


def create_session(config, save_dir):
    """Creates a session to train a model.

    All generated outputs (model checkpoints and logs) will be saved in a directory named after the
    session (the name itself is specified in ``config``), and located in ``save_dir``.

    Args:
        config: a dictionary that provides all required data configuration and trainer parameters; see
            :class:`thelper.train.trainers.Trainer` and :func:`thelper.data.utils.create_loaders` for more information.
            Here, it is only expected to contain a ``name`` field that specifies the name of the session.
        save_dir: the path to the root directory where the session directory should be saved. Note that
            this is not the path to the session directory itself, but its parent, which may also contain
            other session directories.

    .. seealso::
        | :class:`thelper.train.trainers.Trainer`
    """
    logger = thelper.utils.get_func_logger()
    if "name" not in config or not config["name"]:
        raise AssertionError("config missing 'name' field")
    session_name = config["name"]
    logger.info("creating new training session '%s'..." % session_name)
    thelper.utils.setup_globals(config)
    save_dir = thelper.utils.get_save_dir(save_dir, session_name, config)
    logger.debug("session will be saved at '%s'" % os.path.abspath(save_dir).replace("\\", "/"))
    task, train_loader, valid_loader, test_loader = thelper.data.create_loaders(config, save_dir)
    model = thelper.nn.create_model(config, task, save_dir=save_dir)
    loaders = (train_loader, valid_loader, test_loader)
    trainer = thelper.train.create_trainer(session_name, save_dir, config, model, loaders)
    logger.debug("starting trainer")
    if train_loader:
        trainer.train()
    else:
        trainer.eval()
    logger.debug("all done")
    return 0


def resume_session(ckptdata, save_dir, config=None, eval_only=False):
    """Resumes a previously created training session.

    Since the saved checkpoints contain the original session's configuration, the ``config`` argument
    can be set to ``None`` if the session should simply pick up where it was interrupted. Otherwise,
    the ``config`` argument can be set to a new configuration that will override the older one. This is
    useful when fine-tuning a model, or when testing on a new dataset.

    .. warning::
        If a session is resumed with an overriding configuration, the user must make sure that the
        inputs/outputs of the older model are compatible with the new parameters. For example, with
        classifiers, this means that the number of classes parsed by the dataset (and thus to be
        predicted by the model) should remain the same. This is a limitation of the framework that
        should be addressed in a future update.

    .. warning::
        A resumed session will not be compatible with its original RNG states if the number of workers
        used is changed. To get 100% reproducible results, make sure you run with the same worker count.

    Args:
        ckptdata: raw checkpoint data loaded via ``torch.load()``; it will be parsed by the various
            parts of the framework that need to reload their previous state.
        save_dir: the path to the root directory where the session directory should be saved. Note that
            this is not the path to the session directory itself, but its parent, which may also contain
            other session directories.
        config: a dictionary that provides all required data configuration and trainer parameters; see
            :class:`thelper.train.trainers.Trainer` and :func:`thelper.data.utils.create_loaders` for more information.
            Here, it is only expected to contain a ``name`` field that specifies the name of the session.
        eval_only: specifies whether training should be resumed or the model should only be evaluated.

    .. seealso::
        | :class:`thelper.train.trainers.Trainer`
    """
    logger = thelper.utils.get_func_logger()
    if ckptdata is None or not ckptdata:
        raise AssertionError("must provide valid checkpoint data to resume a session!")
    if not config:
        if "config" not in ckptdata or not ckptdata["config"]:
            raise AssertionError("checkpoint data missing 'config' field")
        config = ckptdata["config"]
    if "name" not in config or not config["name"]:
        raise AssertionError("config missing 'name' field")
    session_name = config["name"]
    logger.info("loading training session '%s' objects..." % session_name)
    thelper.utils.setup_globals(config)
    save_dir = thelper.utils.get_save_dir(save_dir, session_name, config, resume=True)
    logger.debug("session will be saved at '%s'" % os.path.abspath(save_dir).replace("\\", "/"))
    new_task, train_loader, valid_loader, test_loader = thelper.data.create_loaders(config, save_dir)
    if "task" not in ckptdata or not ckptdata["task"] or not isinstance(ckptdata["task"], (thelper.tasks.Task, str)):
        raise AssertionError("invalid checkpoint, cannot reload previous model task")
    old_task = thelper.tasks.create_task(ckptdata["task"]) if isinstance(ckptdata["task"], str) else ckptdata["task"]
    if not old_task.check_compat(new_task, exact=True):
        compat_task = None if not old_task.check_compat(new_task) else old_task.get_compat(new_task)
        choice = thelper.utils.query_string("Found discrepancy between old task from checkpoint and new task from config; "+
                                            "which one would you like to resume the session with?\n" +
                                            f"\told: {str(old_task)}\n\tnew: {str(new_task)}\n" +
                                            (f"\tcompat: {str(compat_task)}\n\n" if compat_task is not None else "\n") +
                                            "WARNING: if resuming with new or compat, some weights might be discarded!",
                                            choices=["old", "new", "compat"])
        task = None if choice == "old" else new_task if choice == "new" else compat_task
        if choice != "old":
            # saved optimizer state might cause issues with mismatched tasks, let's get rid of it
            logger.warning("dumping optimizer state to avoid issues when resuming with modified task")
            ckptdata["optimizer"], ckptdata["scheduler"] = None, None
    else:
        task = new_task
    model = thelper.nn.create_model(config, task, save_dir=save_dir, ckptdata=ckptdata)
    loaders = (None if eval_only else train_loader, valid_loader, test_loader)
    trainer = thelper.train.create_trainer(session_name, save_dir, config, model, loaders, ckptdata=ckptdata)
    if eval_only:
        logger.info("evaluating session '%s' checkpoint @ epoch %d" % (trainer.name, trainer.current_epoch))
        trainer.eval()
    else:
        logger.info("resuming training session '%s' @ epoch %d" % (trainer.name, trainer.current_epoch))
        trainer.train()
    logger.debug("all done")
    return 0


def visualize_data(config):
    """Displays the images used in a training session.

    This mode does not generate any output, and is only used to visualize the (transformed) images used
    in a training session. This is useful to debug the data augmentation and base transformation pipelines
    and make sure the modified images are valid. It does not attempt to load a model or instantiate a
    trainer, meaning the related fields are not required inside ``config``.

    If the configuration dictionary includes a 'loaders' field, it will be parsed and used. Otherwise,
    if only a 'datasets' field is available, basic loaders will be instantiated to load the data. The
    'loaders' field can also be ignored if 'viz_ignore_loaders' is found within the config and set
    to ``True``. Each minibatch will be displayed via pyplot. The display will block and wait for user
    input, unless 'viz_block' is found within the config and set to ``False``.

    Args:
        config: a dictionary that provides all required data configuration parameters; see
            :func:`thelper.data.utils.create_loaders` for more information.

    .. seealso::
        | :func:`thelper.data.utils.create_loaders`
        | :func:`thelper.data.utils.create_parsers`
    """
    # todo: move all 'viz' miniconfig stuff to its own section in the config file?
    logger = thelper.utils.get_func_logger()
    logger.info("creating visualization session...")
    thelper.utils.setup_globals(config)
    ignore_loaders = thelper.utils.get_key_def("viz_ignore_loaders", config, default=False)
    use_cv2 = thelper.utils.get_key_def("use_cv2", config, default=True)
    if thelper.utils.get_key_def(["data_config", "loaders"], config, default=None) is None or ignore_loaders:
        datasets, task = thelper.data.create_parsers(config)
        loader_map = {dataset_name: torch.utils.data.DataLoader(dataset,) for dataset_name, dataset in datasets.items()}
        # we assume no transforms were done in the parser, and images are given as read by opencv
        ch_transpose = thelper.utils.get_key_def("viz_ch_transpose", config, False)
        flip_bgr = thelper.utils.get_key_def("viz_flip_bgr", config, False)
    else:
        task, train_loader, valid_loader, test_loader = thelper.data.create_loaders(config)
        loader_map = {"train": train_loader, "valid": valid_loader, "test": test_loader}
        # we assume transforms were done in the loader, and images are given as expected by pytorch
        ch_transpose = thelper.utils.get_key_def("viz_ch_transpose", config, True)
        flip_bgr = thelper.utils.get_key_def("viz_flip_bgr", config, False)
    redraw = None
    block = thelper.utils.get_key_def("viz_block", config, default=True)
    while True:
        choice = thelper.utils.query_string("Which loader would you like to visualize?", choices=list(loader_map.keys()))
        loader = loader_map[choice]
        if loader is None:
            logger.info("loader '%s' is empty" % choice)
            continue
        batch_count = len(loader)
        logger.info("initializing loader '%s' with %d batches..." % (choice, batch_count))
        for batch_idx, samples in enumerate(loader):
            logger.debug("at batch = %d / %d" % (batch_idx, batch_count))
            if "idx" in samples:
                indices = samples["idx"]
                if isinstance(indices, torch.Tensor):
                    logger.debug("(indices = %s)" % indices.tolist())
                else:
                    logger.debug("(indices = %s)" % indices)
            redraw = thelper.utils.draw_minibatch(samples, task, ch_transpose=ch_transpose, flip_bgr=flip_bgr,
                                                  block=block, redraw=redraw, use_cv2=use_cv2)
        logger.info("all done")


def annotate_data(config, save_dir):
    """Launches an annotation session for a dataset using a specialized GUI tool.

    Note that the annotation type must be supported by the GUI tool. The annotations created by the user
    during the session will be saved in the session directory.

    Args:
        config: a dictionary that provides all required dataset and GUI tool configuration parameters; see
            :func:`thelper.data.utils.create_parsers` and :func:`thelper.gui.utils.create_annotator` for more
            information.
        save_dir: the path to the root directory where the session directory should be saved. Note that
            this is not the path to the session directory itself, but its parent, which may also contain
            other session directories.

    .. seealso::
        | :func:`thelper.gui.annotators.Annotator`
        | :func:`thelper.gui.annotators.ImageSegmentAnnotator`
    """
    # import gui here since it imports packages that will cause errors in CLI-only environments
    import thelper.gui
    logger = thelper.utils.get_func_logger()
    if "name" not in config or not config["name"]:
        raise AssertionError("config missing 'name' field")
    session_name = config["name"]
    logger.info("creating annotation session '%s'..." % session_name)
    thelper.utils.setup_globals(config)
    save_dir = thelper.utils.get_save_dir(save_dir, session_name, config)
    logger.debug("session will be saved at '%s'" % os.path.abspath(save_dir).replace("\\", "/"))
    datasets, _ = thelper.data.create_parsers(config)
    annotator = thelper.gui.create_annotator(session_name, save_dir, config, datasets)
    logger.debug("starting annotator")
    annotator.run()
    logger.debug("all done")
    return 0


def main(args=None):
    """Main entrypoint to use with console applications.

    This function parses command line arguments and dispatches the execution based on the selected
    operating mode (new session, resume session, or visualize). Run with ``--help`` for information
    on the available arguments.

    .. warning::
        If you are trying to resume a session that was previously executed using a now unavailable GPU,
        you will have to force the checkpoint data to be loaded on CPU using ``--map-location=cpu`` (or
        using ``-m=cpu``).

    .. seealso::
        | :func:`thelper.cli.create_session`
        | :func:`thelper.cli.resume_session`
        | :func:`thelper.cli.visualize_data`
    """
    ap = argparse.ArgumentParser(description='thelper model trainer application')
    ap.add_argument("--version", default=False, action="store_true", help="prints the version of the library and exits")
    ap.add_argument("-l", "--log", default=None, type=str, help="path to the top-level log file (default: None)")
    ap.add_argument("-v", "--verbose", action="count", default=0, help="set logging terminal verbosity level (additive)")
    ap.add_argument("--force-stdout", action="store_true", default=False, help="force logging output to stdout instead of stderr")
    subparsers = ap.add_subparsers(title="Operating mode", dest="mode")
    new_ap = subparsers.add_parser("new", help="creates a new session from a config file")
    new_ap.add_argument("cfg_path", type=str, help="path to the session configuration file")
    new_ap.add_argument("save_dir", type=str, help="path to the root directory where checkpoints should be saved")
    cl_new_ap = subparsers.add_parser("cl_new", help="creates a new session from a config file for the cluster")
    cl_new_ap.add_argument("cfg_path", type=str, help="path to the session configuration file")
    cl_new_ap.add_argument("save_dir", type=str, help="path to the root directory where checkpoints should be saved")
    resume_ap = subparsers.add_parser("resume", help="resume a session from a checkpoint file")
    resume_ap.add_argument("ckpt_path", type=str, help="path to the checkpoint (or save directory) to resume training from")
    resume_ap.add_argument("-s", "--save-dir", default=None, type=str, help="path to the root directory where checkpoints should be saved")
    resume_ap.add_argument("-m", "--map-location", default=None, help="map location for loading data (default=None)")
    resume_ap.add_argument("-c", "--override-cfg", default=None, help="override config file path (default=None)")
    resume_ap.add_argument("-e", "--eval-only", default=False, action="store_true", help="only run evaluation pass (valid+test)")
    viz_ap = subparsers.add_parser("viz", help="visualize the loaded data for a training/eval session")
    viz_ap.add_argument("cfg_path", type=str, help="path to the session configuration file (or session save directory)")
    annot_ap = subparsers.add_parser("annot", help="launches a dataset annotation session with a GUI tool")
    annot_ap.add_argument("cfg_path", type=str, help="path to the session configuration file (or session save directory)")
    annot_ap.add_argument("save_dir", type=str, help="path to the root directory where annotations should be saved")
    args = ap.parse_args(args=args)
    if args.verbose == 0:
        args.verbose = 3
    if args.verbose > 2:
        log_level = logging.NOTSET
    elif args.verbose > 1:
        log_level = logging.DEBUG
    elif args.verbose == 1:
        log_level = logging.INFO
    else:
        log_level = logging.WARNING
    if args.version:
        print(thelper.__version__)
        return 0
    if args.mode is None:
        ap.print_help()
        return 1
    logging.getLogger().setLevel(logging.NOTSET)
    thelper.logger.propagate = 0
    logger_format = logging.Formatter("[%(asctime)s - %(name)s] %(levelname)s : %(message)s")
    if args.log:
        logger_fh = logging.FileHandler(args.log)
        logger_fh.setLevel(logging.DEBUG)
        logger_fh.setFormatter(logger_format)
        thelper.logger.addHandler(logger_fh)
    stream = sys.stdout if args.force_stdout else None
    logger_ch = logging.StreamHandler(stream=stream)
    logger_ch.setLevel(log_level)
    logger_ch.setFormatter(logger_format)
    thelper.logger.addHandler(logger_ch)
    if args.mode == "new" or args.mode == "cl_new":
        thelper.logger.debug("parsing config at '%s'" % args.cfg_path)
        config = json.load(open(args.cfg_path))
        if args.mode == "cl_new":
            trainer_config = config["trainer"] if "trainer" in config else None
            if trainer_config is not None:
                if ("train_device" in trainer_config or "valid_device" in trainer_config or
                        "test_device" in trainer_config or "device" in trainer_config):
                    raise AssertionError("cannot specify device in config for cluster sessions, it is determined at runtime")
        return create_session(config, args.save_dir)
    elif args.mode == "resume":
        ckptdata = thelper.utils.load_checkpoint(args.ckpt_path, map_location=args.map_location,
                                                 always_load_latest=(not args.eval_only))
        override_config = None
        if args.override_cfg:
            thelper.logger.debug("parsing override config at '%s'" % args.override_cfg)
            override_config = json.load(open(args.override_cfg))
        save_dir = args.save_dir
        if save_dir is None:
            ckpt_dir_path = os.path.dirname(os.path.abspath(args.ckpt_path)) \
                if not os.path.isdir(args.ckpt_path) else os.path.abspath(args.ckpt_path)
            # find session dir by looking for 'logs' directory
            if os.path.isdir(os.path.join(ckpt_dir_path, "logs")):
                save_dir = os.path.abspath(os.path.join(ckpt_dir_path, ".."))
            elif os.path.isdir(os.path.join(ckpt_dir_path, "../logs")):
                save_dir = os.path.abspath(os.path.join(ckpt_dir_path, "../.."))
        return resume_session(ckptdata, save_dir, config=override_config, eval_only=args.eval_only)
    elif args.mode == "viz" or args.mode == "annot":
        if os.path.isdir(args.cfg_path):
            thelper.logger.debug("will search directory '%s' for a config to load..." % args.cfg_path)
            search_cfg_names = ["config.json", "config.latest.json", "test.json"]
            cfg_path = None
            for search_cfg_name in search_cfg_names:
                search_cfg_path = os.path.join(args.cfg_path, search_cfg_name)
                if os.path.isfile(search_cfg_path):
                    cfg_path = search_cfg_path
                    break
            if cfg_path is None or not os.path.isfile(cfg_path):
                raise AssertionError("no valid config found in dir '%s'" % args.cfg_path)
            config = json.load(open(cfg_path))
        else:
            if not os.path.isfile(args.cfg_path):
                raise AssertionError("no config file found at path '%s'" % args.cfg_path)
            config = json.load(open(args.cfg_path))
        if args.mode == "viz":
            return visualize_data(config)
        elif args.mode == "annot":
            return annotate_data(config, args.save_dir)


if __name__ == "__main__":
    main()



"""Schedulers module.

This module contains classes used for scheduling learning rate changes while training a model. All
classes defined here should derive from ``torch.optim.lr_scheduler._LRScheduler`` to remain torch-
compatible.
"""

import bisect
import logging

import torch.nn.functional

logger = logging.getLogger(__name__)


class CustomStepLR(torch.optim.lr_scheduler._LRScheduler):
    """Sets the learning rate of each parameter group using a dictionary of preset scaling factors
    for epoch-based milestones.

    This class can be useful for tuning the learning rate scheduling behavior of a training session
    beyond what is already possible using PyTorch's existing LR scheduler classes.  Note that all
    epoch indices are assumed to be 0-based.

    Usage example in Python::

        # Assuming the optimizer uses lr = 0.05, we hard-code a slow startup...
        # lr = 0.00625   if epoch < 2        (1/8 scale before epoch 2)
        # lr = 0.0125    if 2 <= epoch < 3   (1/4 scale before epoch 3)
        # lr = 0.025     if 3 <= epoch < 4   (1/2 scale before epoch 4)
        # lr = 0.05      if 4 <= epoch < 30  (default scale between epoch 4 and 30)
        # lr = 0.005     if 30 <= epoch < 80 (1/10 scale past epoch 30)
        # lr = 0.0005    if epoch >= 80      (1/100 scale past epoch 80)
        scheduler = CustomStepLR(optimizer, milestones={
            0: 1/8,
            2: 1/4,
            3: 1/2,
            4: 1,
            30: 0.1,
            80: 0.01
        })
        for epoch in range(100):
            scheduler.step(epoch)
            train(...)
            validate(...)

    Usage example inside a session configuration file::

        # ...
        # lists the model optimization parameters for the training session
        "optimization": {
            # lists the optimizer arguments (type, parameters, LR, ...)
            "optimizer": {
                # ...
            },
            # lists the scheduler arguments (field can be omitted if no scheduler is needed)
            "scheduler": {
                # the type used to instantiate the scheduler
                "type": "thelper.optim.schedulers.CustomStepLR",
                # the parameters passed to the scheduler's constructor
                "params": {
                    # by default, the optimizer is passed automatically;
                    # we only need to specify the extra parameters here
                    "milestones": {
                        "1": 1,  # after epoch 1, scale the LR by 1
                        "10": 0.1, # after epoch 10, scale the LR by 0.1
                        "20": 0.01,  # ... and so on
                        "30": 0.001,
                        "40": 0.0001
                    }
                }
            }
        },
        # ...

    Attributes:
        stages: list of epochs where a new scaling factor is to be applied.
        scales: list of scaling factors to apply at each stage.
        milestones: original milestones map provided in the constructor.

    .. seealso::
        | :class:`thelper.train.trainers.Trainer`
    """

    def __init__(self, optimizer, milestones, last_epoch=-1):
        """Receives the optimizer, milestone scaling factor, and initialization state.

        If the milestones do not include the first epoch (idx = 0), then its scaling factor is set
        to 1. When last_epoch is -1, the training is assumed to start from scratch.

        Args:
            optimizer: Wrapped optimizer (PyTorch-compatible object).
            milestones: Map of epoch indices tied to scaling factors. Keys must be increasing.
            last_epoch: The index of last epoch. Default: -1.
        """
        if not isinstance(milestones, dict):
            raise AssertionError("milestones should be provided as a dictionary")
        self.stages = []
        if len(milestones) > 0:
            if isinstance(list(milestones.keys())[0], str):  # fixup for json-based config loaders
                self.stages = [int(key) for key in milestones.keys()]
            elif isinstance(list(milestones.keys())[0], int):
                self.stages = list(milestones.keys())
            else:
                raise AssertionError("milestone stages should be epoch indices (integers)")
            if self.stages != sorted(self.stages):
                raise AssertionError("milestone stages should be increasing integers")
            if not isinstance(list(milestones.values())[0], (float, int)):
                raise AssertionError("milestone scaling factors should be int/float")
        self.scales = [float(scale) for scale in milestones.values()]
        if 0 not in self.stages:
            self.stages.insert(0, int(1))
            self.scales.insert(0, float(1))
        self.milestones = milestones
        super().__init__(optimizer, last_epoch)

    def _get_stage_idx(self, epoch):
        if epoch in self.stages:
            return self.stages.index(epoch)
        return max(bisect.bisect_right(self.stages, epoch) - 1, 0)

    def get_lr(self):
        """Returns the learning rate to use given the current epoch and scaling factors."""
        scale = self.scales[self._get_stage_idx(self.last_epoch)]
        return [base_lr * scale for base_lr in self.base_lrs]



"""Classification task interface module.

This module contains a class that defines the objectives of models/trainers for classification tasks.
"""
import json
import logging
import os
from typing import Optional  # noqa: F401

from thelper.tasks.utils import Task

logger = logging.getLogger(__name__)


class Classification(Task):
    """Interface for input labeling/classification tasks.

    This specialization requests that when given an input tensor, the trained model should
    provide prediction scores for each predefined label (or class). The label names are used
    here to help categorize samples, and to assure that two tasks are only identical when their
    label counts and ordering match.

    Attributes:
        class_names: list of label (class) names to predict (each name should be a string).
        input_key: the key used to fetch input tensors from a sample dictionary.
        label_key: the key used to fetch label (class) names/indices from a sample dictionary.
        meta_keys: the list of extra keys provided by the data parser inside each sample.

    .. seealso::
        | :class:`thelper.tasks.utils.Task`
        | :class:`thelper.train.classif.ImageClassifTrainer`
    """

    def __init__(self, class_names, input_key, label_key, meta_keys=None):
        """Receives and stores the class (or label) names to predict, the input tensor key, the
        groundtruth label (class) key, and the extra (meta) keys produced by the dataset parser(s).

        The class names can be provided as a list of strings, or as a path to a json file that
        contains such a list. The list must contain at least two items. All other arguments are
        used as-is to index dictionaries, and must therefore be key-compatible types.
        """
        super(Classification, self).__init__(input_key, label_key, meta_keys)
        self.class_names = class_names
        if isinstance(class_names, str) and os.path.exists(class_names):
            with open(class_names, "r") as fd:
                self.class_names = json.load(fd)
        if not isinstance(self.class_names, list):
            raise AssertionError("expected class names to be provided as a list")
        if len(self.class_names) < 1:
            raise AssertionError("should have at least one class!")
        if len(self.class_names) != len(set(self.class_names)):
            raise AssertionError("class names should not contain duplicates")

    def get_class_names(self):
        """Returns the list of class names to be predicted by the model."""
        return self.class_names

    def get_nb_classes(self):
        """Returns the number of classes (or labels) to be predicted by the model."""
        return len(self.class_names)

    def get_class_idxs_map(self):
        """Returns the class-label-to-index map used for encoding class labels as integers."""
        return {class_name: idx for idx, class_name in enumerate(self.class_names)}

    def get_class_sizes(self, samples):
        """Given a list of samples, returns a map of sample counts for each class label."""
        class_idxs = self.get_class_sample_map(samples)
        return {class_name: len(class_idxs[class_name]) for class_name in class_idxs}

    def get_class_sample_map(self, samples, unset_key=None):
        """Splits a list of samples based on their labels into a map of sample lists.

        This function is useful if we need to split a dataset based on its label categories in
        order to sort it, augment it, or re-balance it. The samples do not need to be fully loaded
        for this to work, as only their label (gt) value will be queried. If a sample is missing
        its label, it will be ignored and left out of the generated dictionary unless a value is
        given for ``unset_key``.

        Args:
            samples: the samples to split, where each sample is provided as a dictionary.
            unset_key: a key under which all unlabeled samples should be kept (``None`` = ignore).

        Returns:
            A dictionary that maps each class label to its corresponding list of samples.
        """
        if samples is None or not samples:
            raise AssertionError("provided invalid sample list")
        sample_idxs = {class_name: [] for class_name in self.class_names}
        if unset_key is not None and not isinstance(unset_key, str):
            raise AssertionError("unset class name key should be string, just like other class names")
        elif unset_key in sample_idxs:
            raise AssertionError("unset class name key already in class names list")
        else:
            sample_idxs[unset_key] = []
        label_key = self.get_gt_key()
        for sample_idx, sample in enumerate(samples):
            if label_key is None or label_key not in sample:
                if unset_key is not None:
                    class_name = unset_key
                else:
                    continue
            else:
                class_name = sample[label_key]
                if isinstance(class_name, str):
                    if class_name not in self.class_names:
                        raise AssertionError("label '%s' not found in class names provided earlier" % class_name)
                elif isinstance(class_name, int):
                    # dataset must already be using indices, we will forgive this...
                    # (this is pretty much always the case for torchvision datasets)
                    if class_name < 0 or class_name >= len(self.class_names):
                        raise AssertionError("class name given as out-of-range index (%d) for class list" % class_name)
                    class_name = self.class_names[class_name]
                else:
                    raise AssertionError("unexpected sample label type (need string!)")
            sample_idxs[class_name].append(sample_idx)
        return sample_idxs

    def check_compat(self, other, exact=False):
        # type: (Classification, Optional[bool]) -> bool
        """Returns whether the current task is compatible with the provided one or not.

        This is useful for sanity-checking, and to see if the inputs/outputs of two models
        are compatible. If ``exact = True``, all fields will be checked for exact (perfect)
        compatibility (in this case, matching meta keys and class name order).
        """
        if isinstance(other, Classification):
            # if both tasks are related to classification, gt keys and class names must match
            return (self.get_input_key() == other.get_input_key() and
                    (self.get_gt_key() is None or other.get_gt_key() is None or
                     self.get_gt_key() == other.get_gt_key()) and
                    all([cls in self.get_class_names() for cls in other.get_class_names()]) and
                    (not exact or (self.get_class_names() == other.get_class_names() and
                                   self.get_meta_keys() == other.get_meta_keys())))
        elif type(other) == Task:
            # if 'other' simply has no gt, compatibility rests on input key only
            return not exact and self.get_input_key() == other.get_input_key() and other.get_gt_key() is None
        return False

    def get_compat(self, other):
        """Returns a task instance compatible with the current task and the given one."""
        if isinstance(other, Classification):
            if self.get_input_key() != other.get_input_key():
                raise AssertionError("input key mismatch, cannot create compatible task")
            if self.get_gt_key() is not None \
                    and other.get_gt_key() is not None \
                    and self.get_gt_key() != other.get_gt_key():
                raise AssertionError("gt key mismatch, cannot create compatible task")
            meta_keys = list(set(self.get_meta_keys() + other.get_meta_keys()))
            # cannot use set for class names, order needs to stay intact!
            other_names = [name for name in other.get_class_names() if name not in self.get_class_names()]
            class_names = self.get_class_names() + other_names
            return Classification(class_names, self.get_input_key(), self.get_gt_key(), meta_keys=meta_keys)
        elif type(other) == Task:
            if not self.check_compat(other):
                raise AssertionError("cannot create compatible task instance between:\n"
                                     "\tself: %s\n\tother: %s" % (str(self), str(other)))
            meta_keys = list(set(self.get_meta_keys() + other.get_meta_keys()))
            return Classification(self.get_class_names(), self.get_input_key(), self.get_gt_key(), meta_keys=meta_keys)
        else:
            raise AssertionError("cannot combine task type '%s' with '%s'"
                                 % (str(other.__class__), str(self.__class__)))

    def __repr__(self):
        """Creates a print-friendly representation of a classification task."""
        return self.__class__.__module__ + "." + self.__class__.__qualname__ + ": " + str({
            "class_names": self.get_class_names(),
            "input_key": self.get_input_key(),
            "label_key": self.get_gt_key(),
            "meta_keys": self.get_meta_keys(),
        })



# noinspection PyPackageRequirements
import mock


def test_main_version():
    from thelper.cli import main
    main(["--version"])


def test_main_init():
    import thelper.__main__
    import thelper.cli
    with mock.patch.object(thelper.cli, "main", return_value=1337):
        with mock.patch.object(thelper.__main__, "__name__", "__main__"):
            with mock.patch.object(thelper.__main__.sys, "exit") as mock_exit:
                thelper.__main__.init()
                assert mock_exit.call_args[0][0] == 1337



"""Optimization/metrics utility module.

This module contains utility functions and tools used by the other modules of this package.
"""
import copy
import inspect
import logging

import torch

import thelper.optim.metrics
import thelper.utils

logger = logging.getLogger(__name__)


def create_loss_fn(config, model, loader=None, uploader=None):
    """Instantiates and returns the loss function to use for training.

    The default way to specify the loss function to use is to provide a callable type to instantiate
    as well as its initialization parameters. For example::

        # ...
        "loss": {
            # if we want to use PyTorch's cross entropy loss:
            #  >>> loss_fn = torch.nn.CrossEntropyLoss(**params)
            #  >>> ...
            #  >>> loss = loss_fn(pred, target)
            "type": "torch.nn.CrossEntropyLoss"
            "params": {
                "weight": [ ... ],
                "reduction": "mean",
                # ...
            }
        },
        # ...

    The loss function can also be queried from a member function of the model class, as such::

        # ...
        "loss": {
            # to query the model for the loss function:
            #  >>> loss_fn = model.get_loss_fn(**params)
            #  >>> ...
            #  >>> loss = loss_fn(pred, target)
            "model_getter": "get_loss_fn"
            "params": {
                # ...
            }
        },
        # ...

    If the model is supposed to compute its own loss, we suggest creating a specialized trainer class. In that
    case only, the 'loss' field can be omitted from the session configuration file.

    If the task is related to image classification or semantic segmentation, the classes can be weighted based
    on extra parameters in the loss configuration. The strategy used to compute the weights is related to the
    one in :class:`thelper.data.samplers.WeightedSubsetRandomSampler`. The exact parameters that are expected
    for class reweighting are the following:

    - ``weight_distribution`` (mandatory, toggle): the dictionary of weights assigned to each class, or the
      rebalancing strategy to use. If omitted entirely, no class weighting will be performed.
    - ``weight_param_name`` (optional, default="weight"): name of the constructor parameter that expects the
      weight list.
    - ``weight_max`` (optional, default=inf): the maximum weight that can be assigned to a class.
    - ``weight_min`` (optional, default=0): the minimum weight that can be assigned to a class.
    - ``weight_norm`` (optional, default=True): specifies whether the weights should be normalized or not.

    This function also supports an extra special parameter if the task is related to semantic segmentation:
    ``ignore_index``. If this parameter is found and not ``None`` (integer), then the loss function will ignore
    the given value when computing the loss of a sample. The exact parameters that are expected in this case are
    the following:

    - ``ignore_index_param_name`` (optional, default="ignore_index"): name of the constructor parameter that expects
      the ignore value.
    - ``ignore_index_label_name`` (optional, default="dontcare"): name of the label to pass the ignore value from.

    """
    # todo: add flag to toggle loss comp in validation? (add to trainer config maybe?)
    logger.debug("loading loss function")
    if isinstance(model, torch.nn.DataParallel):
        model = model.module  # to avoid interface getter issues
    if uploader is None:
        def uploader(x):
            return x
    if not isinstance(config, dict):
        raise AssertionError("config should be provided as a dictionary")
    if "model_getter" in config and "type" in config:
        raise AssertionError("loss config cannot have both 'model_attrib_name' and 'type' fields")
    if "model_getter" in config:
        model_getter_name = config["model_getter"]
        if not isinstance(model_getter_name, str):
            raise AssertionError("unexpected model getter name type")
        if not hasattr(model, model_getter_name) or not callable(getattr(model, model_getter_name)):
            raise AssertionError("invalid model getter attribute")
        loss_type = getattr(model, model_getter_name)
    elif "type" in config:
        loss_type = thelper.utils.import_class(config["type"])
    else:
        raise AssertionError("loss config missing 'type' or 'model_attrib_name' field")
    loss_params = thelper.utils.get_key_def(["params", "parameters"], config, {})
    loss_params = copy.deepcopy(loss_params)  # required here, we might add some parameters below
    if thelper.utils.str2bool(thelper.utils.get_key_def("weight_classes", config, False)) or \
       thelper.utils.get_key_def("weight_distribution", config, None) is not None:
        if not thelper.utils.str2bool(thelper.utils.get_key_def("weight_classes", config, True)):
            raise AssertionError("'weight_classes' now deprecated, set 'weight_distribution' directly to toggle on")
        if not isinstance(model.task, (thelper.tasks.Classification, thelper.tasks.Segmentation)):
            raise AssertionError("task type does not support class weighting")
        weight_param_name = "weight"
        if "weight_param_name" in config:
            weight_param_name = config["weight_param_name"]
        if "weight_distribution" not in config:
            raise AssertionError("missing 'weight_distribution' field in loss config")
        weight_distrib = config["weight_distribution"]
        if isinstance(weight_distrib, dict):
            for label, weight in weight_distrib.items():
                if label not in model.task.get_class_names():
                    raise AssertionError("weight distribution label '%s' not in dataset class list" % label)
                if not isinstance(weight, float):
                    raise AssertionError("expected weight distrib map to provide weights as floats directly")
        elif isinstance(weight_distrib, str):
            weight_max = float("inf")
            if "weight_max" in config:
                weight_max = float(config["weight_max"])
            weight_min = 0
            if "weight_min" in config:
                weight_min = float(config["weight_min"])
            if weight_distrib != "uniform":
                if loader is None or not loader:
                    raise AssertionError("cannot get class sizes, no training data available")
                label_sizes_map = model.task.get_class_sizes(loader.dataset)
                weight_norm = False
            else:
                label_sizes_map = {label: -1 for label in model.task.get_class_names()}  # counts don't matter
                weight_norm = True
            if "weight_norm" in config:
                weight_norm = thelper.utils.str2bool(config["weight_norm"])
            weight_distrib = thelper.data.utils.get_class_weights(label_sizes_map, weight_distrib, invmax=True,
                                                                  maxw=weight_max, minw=weight_min, norm=weight_norm)
        else:
            raise AssertionError("unexpected weight distribution strategy (should be map or string)")
        weight_list_str = "weight_distribution: {"
        for label, weight in weight_distrib.items():
            weight_list_str += "\n  \"%s\": %s," % (label, weight)
        logger.info(weight_list_str + "\n}")
        weight_list = [weight_distrib[label] if label in weight_distrib else 1.0 for label in model.task.get_class_names()]
        loss_params[weight_param_name] = uploader(torch.FloatTensor(weight_list))
    if isinstance(model.task, thelper.tasks.Segmentation):
        ignore_index_param_name = thelper.utils.get_key_def("ignore_index_param_name", config, "ignore_index")
        ignore_index_label_name = thelper.utils.get_key_def("ignore_index_label_name", config, "dontcare")
        loss_sig = inspect.signature(loss_type)
        if ignore_index_param_name in loss_sig.parameters:
            if ignore_index_label_name != "dontcare":
                loss_params[ignore_index_param_name] = model.task.get_class_idxs_map()[ignore_index_label_name]
            else:
                loss_params[ignore_index_param_name] = model.task.get_dontcare_val()
    loss = loss_type(**loss_params)
    return loss


def create_metrics(config):
    """Instantiates and returns the metrics defined in the configuration dictionary.

    All arguments are expected to be handed in through the configuration via a dictionary named 'params'.
    """
    if not isinstance(config, dict):
        raise AssertionError("config should be provided as a dictionary")
    metrics = {}
    for name, metric_config in config.items():
        if not isinstance(metric_config, dict):
            raise AssertionError("metric config should be provided as a dictionary")
        if "type" not in metric_config or not metric_config["type"]:
            raise AssertionError("metric config missing 'type' field")
        metric_type = thelper.utils.import_class(metric_config["type"])
        metric_params = thelper.utils.get_key_def(["params", "parameters"], metric_config, {})
        metric = metric_type(**metric_params)
        if not isinstance(metric, thelper.optim.metrics.Metric):
            raise AssertionError("invalid metric type, must derive from 'thelper.optim.metrics.Metric'")
        goal = getattr(metric, "goal", None)
        if not callable(goal):
            raise AssertionError("expected metric to define 'goal' based on parent interface")
        metrics[name] = metric
    return metrics


def create_optimizer(config, model):
    """Instantiates and returns the optimizer to use for training.

    By default, the optimizer will be instantiated with the model parameters given as the first argument
    of its constructor. All supplementary arguments are expected to be handed in through the configuration
    via a dictionary named 'params'.
    """
    logger.debug("loading optimizer")
    if not isinstance(config, dict):
        raise AssertionError("config should be provided as a dictionary")
    if "type" not in config or not config["type"]:
        raise AssertionError("optimizer config missing 'type' field")
    optimizer_type = thelper.utils.import_class(config["type"])
    optimizer_params = thelper.utils.get_key_def(["params", "parameters"], config, {})
    optimizer = optimizer_type(filter(lambda p: p.requires_grad, model.parameters()), **optimizer_params)
    return optimizer


def create_scheduler(config, optimizer):
    """Instantiates and returns the learning rate scheduler to use for training.

    All arguments are expected to be handed in through the configuration via a dictionary named 'params'.
    """
    logger.debug("loading scheduler")
    if not isinstance(config, dict):
        raise AssertionError("config should be provided as a dictionary")
    if "type" not in config or not config["type"]:
        raise AssertionError("scheduler config missing 'type' field")
    scheduler_type = thelper.utils.import_class(config["type"])
    scheduler_params = thelper.utils.get_key_def(["params", "parameters"], config, {})
    scheduler = scheduler_type(optimizer, **scheduler_params)
    scheduler_step_metric = None
    if "step_metric" in config:
        scheduler_step_metric = config["step_metric"]
    return scheduler, scheduler_step_metric


def get_lr(optimizer):
    """Returns the optimizer's learning rate, or 0 if not found."""
    for param_group in optimizer.param_groups:
        if "lr" in param_group:
            return param_group["lr"]
    return 0



"""Graphical User Interface (GUI) package.

This package contains various tools and annotators used to simplify user interactions with data
or models. Since the training framework is CLI-based, these tools are not used to train models,
but can be helpful when debugging them. They can also be used to annotate and explore datasets.
"""

import logging

import thelper.gui.annotators  # noqa: F401
import thelper.gui.utils  # noqa: F401
from thelper.gui.annotators import Annotator  # noqa: F401
from thelper.gui.annotators import ImageSegmentAnnotator  # noqa: F401
from thelper.gui.utils import create_annotator  # noqa: F401

logger = logging.getLogger("thelper.gui")



"""Top-level package for the 'thelper' framework.

Running ``import thelper`` will recursively import all important subpackages and modules.
"""

import logging

import thelper.cli  # noqa: F401
import thelper.data  # noqa: F401
import thelper.nn  # noqa: F401
import thelper.optim  # noqa: F401
import thelper.tasks  # noqa: F401
import thelper.train  # noqa: F401
import thelper.transforms  # noqa: F401
import thelper.typedefs  # noqa: F401
import thelper.utils  # noqa: F401

logger = logging.getLogger("thelper")

__url__ = "https://github.com/plstcharles/thelper"
__version__ = "0.2.8"



# InceptionResNetv2 derived from:
# https://github.com/Cadene/pretrained-models.pytorch/blob/master/pretrainedmodels/models/inceptionresnetv2.py
import torch
import torch.nn

import thelper.nn
import thelper.nn.coordconv


class BasicConv2d(torch.nn.Module):

    def __init__(self, in_planes, out_planes, kernel_size, stride, padding=0):
        super().__init__()
        self.conv = torch.nn.Conv2d(in_planes, out_planes,
                                    kernel_size=kernel_size, stride=stride,
                                    padding=padding, bias=False)  # verify bias false
        self.bn = torch.nn.BatchNorm2d(out_planes,
                                       eps=0.001,  # value found in tensorflow
                                       momentum=0.1,  # default pytorch value
                                       affine=True)
        self.relu = torch.nn.ReLU(inplace=False)

    def forward(self, x):
        x = self.conv(x)
        x = self.bn(x)
        x = self.relu(x)
        return x


class Mixed_5b(torch.nn.Module):

    def __init__(self):
        super().__init__()
        self.branch0 = BasicConv2d(192, 96, kernel_size=1, stride=1)
        self.branch1 = torch.nn.Sequential(
            BasicConv2d(192, 48, kernel_size=1, stride=1),
            BasicConv2d(48, 64, kernel_size=5, stride=1, padding=2)
        )
        self.branch2 = torch.nn.Sequential(
            BasicConv2d(192, 64, kernel_size=1, stride=1),
            BasicConv2d(64, 96, kernel_size=3, stride=1, padding=1),
            BasicConv2d(96, 96, kernel_size=3, stride=1, padding=1)
        )
        self.branch3 = torch.nn.Sequential(
            torch.nn.AvgPool2d(3, stride=1, padding=1, count_include_pad=False),
            BasicConv2d(192, 64, kernel_size=1, stride=1)
        )

    def forward(self, x):
        x0 = self.branch0(x)
        x1 = self.branch1(x)
        x2 = self.branch2(x)
        x3 = self.branch3(x)
        out = torch.cat((x0, x1, x2, x3), 1)
        return out


class Block35(torch.nn.Module):

    def __init__(self, scale=1.0):
        super().__init__()
        self.scale = scale
        self.branch0 = BasicConv2d(320, 32, kernel_size=1, stride=1)
        self.branch1 = torch.nn.Sequential(
            BasicConv2d(320, 32, kernel_size=1, stride=1),
            BasicConv2d(32, 32, kernel_size=3, stride=1, padding=1)
        )
        self.branch2 = torch.nn.Sequential(
            BasicConv2d(320, 32, kernel_size=1, stride=1),
            BasicConv2d(32, 48, kernel_size=3, stride=1, padding=1),
            BasicConv2d(48, 64, kernel_size=3, stride=1, padding=1)
        )
        self.conv2d = torch.nn.Conv2d(128, 320, kernel_size=1, stride=1)
        self.relu = torch.nn.ReLU(inplace=False)

    def forward(self, x):
        x0 = self.branch0(x)
        x1 = self.branch1(x)
        x2 = self.branch2(x)
        out = torch.cat((x0, x1, x2), 1)
        out = self.conv2d(out)
        out = out * self.scale + x
        out = self.relu(out)
        return out


class Mixed_6a(torch.nn.Module):

    def __init__(self):
        super().__init__()
        self.branch0 = BasicConv2d(320, 384, kernel_size=3, stride=2)
        self.branch1 = torch.nn.Sequential(
            BasicConv2d(320, 256, kernel_size=1, stride=1),
            BasicConv2d(256, 256, kernel_size=3, stride=1, padding=1),
            BasicConv2d(256, 384, kernel_size=3, stride=2)
        )
        self.branch2 = torch.nn.MaxPool2d(3, stride=2)

    def forward(self, x):
        x0 = self.branch0(x)
        x1 = self.branch1(x)
        x2 = self.branch2(x)
        out = torch.cat((x0, x1, x2), 1)
        return out


class Block17(torch.nn.Module):

    def __init__(self, scale=1.0):
        super().__init__()
        self.scale = scale
        self.branch0 = BasicConv2d(1088, 192, kernel_size=1, stride=1)
        self.branch1 = torch.nn.Sequential(
            BasicConv2d(1088, 128, kernel_size=1, stride=1),
            BasicConv2d(128, 160, kernel_size=(1, 7), stride=1, padding=(0, 3)),
            BasicConv2d(160, 192, kernel_size=(7, 1), stride=1, padding=(3, 0))
        )
        self.conv2d = torch.nn.Conv2d(384, 1088, kernel_size=1, stride=1)
        self.relu = torch.nn.ReLU(inplace=False)

    def forward(self, x):
        x0 = self.branch0(x)
        x1 = self.branch1(x)
        out = torch.cat((x0, x1), 1)
        out = self.conv2d(out)
        out = out * self.scale + x
        out = self.relu(out)
        return out


class Mixed_7a(torch.nn.Module):

    def __init__(self):
        super().__init__()
        self.branch0 = torch.nn.Sequential(
            BasicConv2d(1088, 256, kernel_size=1, stride=1),
            BasicConv2d(256, 384, kernel_size=3, stride=2)
        )
        self.branch1 = torch.nn.Sequential(
            BasicConv2d(1088, 256, kernel_size=1, stride=1),
            BasicConv2d(256, 288, kernel_size=3, stride=2)
        )
        self.branch2 = torch.nn.Sequential(
            BasicConv2d(1088, 256, kernel_size=1, stride=1),
            BasicConv2d(256, 288, kernel_size=3, stride=1, padding=1),
            BasicConv2d(288, 320, kernel_size=3, stride=2)
        )
        self.branch3 = torch.nn.MaxPool2d(3, stride=2)

    def forward(self, x):
        x0 = self.branch0(x)
        x1 = self.branch1(x)
        x2 = self.branch2(x)
        x3 = self.branch3(x)
        out = torch.cat((x0, x1, x2, x3), 1)
        return out


class Block8(torch.nn.Module):

    def __init__(self, scale=1.0, noReLU=False):
        super(Block8, self).__init__()
        self.scale = scale
        self.noReLU = noReLU
        self.branch0 = BasicConv2d(2080, 192, kernel_size=1, stride=1)
        self.branch1 = torch.nn.Sequential(
            BasicConv2d(2080, 192, kernel_size=1, stride=1),
            BasicConv2d(192, 224, kernel_size=(1, 3), stride=1, padding=(0, 1)),
            BasicConv2d(224, 256, kernel_size=(3, 1), stride=1, padding=(1, 0))
        )
        self.conv2d = torch.nn.Conv2d(448, 2080, kernel_size=1, stride=1)
        if not self.noReLU:
            self.relu = torch.nn.ReLU(inplace=False)

    def forward(self, x):
        x0 = self.branch0(x)
        x1 = self.branch1(x)
        out = torch.cat((x0, x1), 1)
        out = self.conv2d(out)
        out = out * self.scale + x
        if not self.noReLU:
            out = self.relu(out)
        return out


class InceptionResNetV2(thelper.nn.Module):

    def __init__(self, task, input_channels=3):
        super().__init__(task)
        self.conv2d_1a = BasicConv2d(input_channels, 32, kernel_size=3, stride=2)
        self.conv2d_2a = BasicConv2d(32, 32, kernel_size=3, stride=1)
        self.conv2d_2b = BasicConv2d(32, 64, kernel_size=3, stride=1, padding=1)
        self.maxpool_3a = torch.nn.MaxPool2d(3, stride=2)
        self.conv2d_3b = BasicConv2d(64, 80, kernel_size=1, stride=1)
        self.conv2d_4a = BasicConv2d(80, 192, kernel_size=3, stride=1)
        self.maxpool_5a = torch.nn.MaxPool2d(3, stride=2)
        self.mixed_5b = Mixed_5b()
        self.repeat = torch.nn.Sequential(
            Block35(scale=0.17),
            Block35(scale=0.17),
            Block35(scale=0.17),
            Block35(scale=0.17),
            Block35(scale=0.17),
            Block35(scale=0.17),
            Block35(scale=0.17),
            Block35(scale=0.17),
            Block35(scale=0.17),
            Block35(scale=0.17)
        )
        self.mixed_6a = Mixed_6a()
        self.repeat_1 = torch.nn.Sequential(
            Block17(scale=0.10),
            Block17(scale=0.10),
            Block17(scale=0.10),
            Block17(scale=0.10),
            Block17(scale=0.10),
            Block17(scale=0.10),
            Block17(scale=0.10),
            Block17(scale=0.10),
            Block17(scale=0.10),
            Block17(scale=0.10),
            Block17(scale=0.10),
            Block17(scale=0.10),
            Block17(scale=0.10),
            Block17(scale=0.10),
            Block17(scale=0.10),
            Block17(scale=0.10),
            Block17(scale=0.10),
            Block17(scale=0.10),
            Block17(scale=0.10),
            Block17(scale=0.10)
        )
        self.mixed_7a = Mixed_7a()
        self.repeat_2 = torch.nn.Sequential(
            Block8(scale=0.20),
            Block8(scale=0.20),
            Block8(scale=0.20),
            Block8(scale=0.20),
            Block8(scale=0.20),
            Block8(scale=0.20),
            Block8(scale=0.20),
            Block8(scale=0.20),
            Block8(scale=0.20)
        )
        self.block8 = Block8(noReLU=True)
        self.conv2d_7b = BasicConv2d(2080, 1536, kernel_size=1, stride=1)
        self.avgpool_1a = torch.nn.AvgPool2d(8, count_include_pad=False)
        self.last_linear = torch.nn.Linear(1536, 1000)
        self.set_task(task)

    def features(self, input):
        x = self.conv2d_1a(input)
        x = self.conv2d_2a(x)
        x = self.conv2d_2b(x)
        x = self.maxpool_3a(x)
        x = self.conv2d_3b(x)
        x = self.conv2d_4a(x)
        x = self.maxpool_5a(x)
        x = self.mixed_5b(x)
        x = self.repeat(x)
        x = self.mixed_6a(x)
        x = self.repeat_1(x)
        x = self.mixed_7a(x)
        x = self.repeat_2(x)
        x = self.block8(x)
        x = self.conv2d_7b(x)
        return x

    def logits(self, features):
        x = self.avgpool_1a(features)
        x = x.view(x.size(0), -1)
        x = self.last_linear(x)
        return x

    def forward(self, input):
        x = self.features(input)
        x = self.logits(x)
        return x

    def set_task(self, task):
        if isinstance(task, thelper.tasks.Classification):
            num_classes = len(task.get_class_names())
            if self.last_linear.out_features != num_classes:
                self.last_linear = torch.nn.Linear(1536, num_classes)
        else:
            raise AssertionError("missing impl for non-classif task type")
        self.task = task
